import copy
import math
from typing import List, Any, Union, Dict

from golem.utilities.requirements_notificator import warn_requirement

try:
    import torch
except ModuleNotFoundError:
    warn_requirement('torch', 'other_requirements/adaptive.txt')

import numpy as np
from mabwiser.mab import MAB, LearningPolicy, NeighborhoodPolicy
from mabwiser.utils import Arm, Constants, Num

from golem.core.log import default_log

import warnings

warnings.filterwarnings("ignore")


class NeuralMAB(MAB):
    """
    Neural Multi-Armed Bandit.
    The main concept is explained in the article: https://arxiv.org/abs/2012.01780.
    Deep representation is formed with NN and Contextual Multi-Armed Bandit is integrated to choose arm.

    NB! Neural MABs can be used with 1.8.0 version of torch since some methods are deprecated in later versions,
    however, python of version 3.10 is not supported in this version of torch.
    """

    def __init__(self, arms: List[Arm],
                 learning_policy: Any = LearningPolicy.UCB1(alpha=1.25),
                 neighborhood_policy: Any = NeighborhoodPolicy.Clusters(),
                 seed: int = Constants.default_seed,
                 n_jobs: int = 1):

        super().__init__(arms, learning_policy, neighborhood_policy, seed, n_jobs)
        self.nn_with_se = NNWithShallowExploration(context_size=1, arms_count=len(arms))
        self.arms = arms
        self.seed = seed
        self.n_jobs = n_jobs
        self.log = default_log('NeuralMAB')
        # to track when GNN needs to be updated
        self.iter = 0
        self._indices = list(range(len(arms)))
        self._mab = MAB(arms=self._indices,
                        learning_policy=learning_policy,
                        neighborhood_policy=neighborhood_policy,
                        n_jobs=n_jobs)
        self.is_fitted = False

    def _initial_fit_mab(self, context: Any):
        """ Initial fit for Contextual Multi-Armed Bandit.
        At this step, all hands are assigned the same weights with the very first context
        that is fed to the bandit. """
        # initial fit for mab
        n = len(self.arms)
        uniform_rewards = [1. / n] * n
        deep_context = self._get_deep_context(context=context)
        self._mab.fit(decisions=self._indices, rewards=uniform_rewards, contexts=n * [deep_context])
        self.is_fitted = True

    def partial_fit(self, decisions: List[Any], rewards: List[float], contexts: List[Any] = None):

        # get deep contexts, calculate regret and update weights for NN (once in _H_q iters)
        deep_contexts = self.nn_with_se.partial_fit(iter=self.iter, decisions=decisions,
                                                    rewards=rewards, contexts=contexts)
        self.iter += 1

        # update contextual mab with deep contexts
        self._mab.partial_fit(decisions=decisions, contexts=deep_contexts, rewards=rewards)

    def predict(self, contexts: Any = None) -> Union[Arm, List[Arm]]:
        """ Predicts which arm to pull to get maximum reward. """
        if not self.is_fitted:
            self._initial_fit_mab(context=contexts)
        deep_context = self._get_deep_context(context=contexts)
        a_choose = self._mab.predict(contexts=[deep_context])
        return a_choose

    def predict_expectations(self, contexts: Any = None) -> Union[Dict[Arm, Num], List[Dict[Arm, Num]]]:
        """ Returns expected reward for each arm. """
        if not self.is_fitted:
            self._initial_fit_mab(context=contexts)
        deep_context = self._get_deep_context(context=contexts)
        return self._mab.predict_expectations(contexts=[deep_context])

    def _get_deep_context(self, context: Any) -> List[Any]:
        """ Returns deep representation of context. """
        temp = self.nn_with_se.transfer(context, 0, len(self.arms))
        feat = self.nn_with_se.feature_extractor(temp, self.nn_with_se.W).numpy().squeeze()
        return list(feat)


class NNWithShallowExploration:
    """ Neural Network with shallow exploration which means that weights are updated every _H_q iterations. """

    def __init__(self, context_size: int, arms_count: int):
        """
        Initial fit for NN.
        beta -- parameter for UCB exploration
        H_q -- how many time steps to update NN
        interT -- internal steps for GD
        """
        self._beta = 0.02
        self._lambd = 1
        self._lr = 0.001
        self._H_q = 5
        self._interT = 1000
        self._hidden_dim = [1000, 1000]
        hid_dim_lst = self._hidden_dim
        dim_second_last = self._hidden_dim[-1] * 2

        dim_for_init = [context_size + arms_count] + hid_dim_lst + [1]
        self.arms_count = arms_count
        self.W0, total_dim = self._initialization(dim_for_init)
        self.LAMBDA = self._lambd * torch.eye(dim_second_last, dtype=torch.double)
        self.bb = torch.zeros(self.LAMBDA.size()[0], dtype=torch.double).view(-1, 1)

        theta = np.random.randn(dim_second_last, 1) / np.sqrt(dim_second_last)
        self.theta = torch.from_numpy(theta)

        self.THETA_action = torch.tensor([])
        self.CONTEXT_action = torch.tensor([])
        self.REWARD_action = torch.tensor([])
        self.result_neuralucb = []
        self.W = copy.deepcopy(self.W0)
        self.summ = 0
        self.log = default_log('NNWithShallowExploration')

    def partial_fit(self, iter: int,
                    decisions: List[Any], rewards: List[float], contexts: List[Any] = None):
        deep_contexts = []

        # update NN and calculate reward
        for decision, context, reward in zip(decisions, contexts, rewards):

            # calculate reward
            temp = self.transfer(context, decision, self.arms_count)
            feat = self.feature_extractor(temp, self.W)
            deep_contexts.append(list(feat.numpy().squeeze()))
            expected_reward = torch.mm(self.theta.view(1, -1), feat) + self._beta * self.UCB(self.LAMBDA, feat)

            self.summ += (expected_reward - reward)
            self.result_neuralucb.append(self.summ)

            # gather dataset for next NN training (context_action and reward_action)
            if np.mod(iter, self._H_q) == 0:
                context_action = temp
                reward_action = torch.tensor([reward], dtype=torch.double)
            else:
                context_action = torch.cat((self.CONTEXT_action, temp), 1)
                reward_action = torch.cat((self.REWARD_action, torch.tensor([reward], dtype=torch.double)), 0)

            # update LAMBDA and bb
            self.LAMBDA += torch.mm(self.feature_extractor(temp, self.W),
                                    self.feature_extractor(temp, self.W).t())
            self.bb += reward * self.feature_extractor(temp, self.W)
            theta, _ = torch.solve(self.bb, self.LAMBDA)

            if np.mod(iter, self._H_q) == 0:
                theta_action = theta.view(-1, 1)
            else:
                theta_action = torch.cat((self.THETA_action, theta.view(-1, 1)), 1)

        # update weight of NN
        if np.mod(iter, self._H_q) == self._H_q - 1:
            self.log.info(f'Current regret: {self.summ}')
            self.W = self.train_with_shallow_exploration(context_action, reward_action, self.W0,
                                                         self._interT, self._lr, theta_action, self._H_q)
        return deep_contexts

    @staticmethod
    def UCB(A, phi):
        """ Ucb term. """
        try:
            tmp, _ = torch.solve(phi, A)
        except Exception:
            tmp = torch.Tensor(np.linalg.solve(A, phi))

        return torch.sqrt(torch.mm(torch.transpose(phi, 0, 1).double(), tmp.double()))

    @staticmethod
    def transfer(c, a, arm_size):
        """
        Transfer an array context + action to new context with dimension 2*(__context__ + __armsize__).
        """
        action = np.zeros(arm_size)
        action[a] = 1
        c_final = np.append(c, action)
        c_final = torch.from_numpy(c_final)
        c_final = c_final.view((len(c_final), 1))
        c_final = c_final.repeat(2, 1)
        return c_final

    def train_with_shallow_exploration(self, X, Y, W_start, T, et, THETA, H):
        """ Gd-based model training with shallow exploration
        Dataset X, label Y. """
        W = copy.deepcopy(W_start)
        X = X[:, -H:]
        Y = Y[-H:]
        THETA = THETA[:, -H:]

        prev_loss = 1000000
        prev_loss_1k = 1000000
        for i in range(0, T):
            grad = self._gradient_loss(X, Y, W, THETA)
            for j in range(0, len(W) - 1):
                W[j] = W[j] - et * grad[j]

            curr_loss = self._loss(X, Y, W, THETA)
            if i % 100 == 0:
                print('------', curr_loss)
                if curr_loss > prev_loss_1k:
                    et = et * 0.1
                    print('lr/10 to', et)

                prev_loss_1k = curr_loss

            # early stopping
            if abs(curr_loss - prev_loss) < 1e-6:
                break
            prev_loss = curr_loss
        return W

    @staticmethod
    def _initialization(dim):
        """ Initialization.
        dim consists of (d1, d2,...), where dl = 1 (placeholder, deprecated). """
        w = []
        total_dim = 0
        for i in range(0, len(dim) - 1):
            if i < len(dim) - 2:
                temp = np.random.randn(dim[i + 1], dim[i]) / np.sqrt(dim[i + 1])
                temp = np.kron(np.eye(2, dtype=int), temp)
                temp = torch.from_numpy(temp)
                w.append(temp)
                total_dim += dim[i + 1] * dim[i] * 4
            else:
                temp = np.random.randn(dim[i + 1], dim[i]) / np.sqrt(dim[i])
                temp = np.kron([[1, -1]], temp)
                temp = torch.from_numpy(temp)
                w.append(temp)
                total_dim += dim[i + 1] * dim[i] * 2

        return w, total_dim

    @staticmethod
    def feature_extractor(x, W):
        """ Functions feature extractor.
        x is the input, dimension is d; W is the list of parameter matrices. """
        depth = len(W)
        output = x
        for i in range(0, depth - 1):
            output = torch.mm(W[i], output)
            output = output.clamp(min=0)

        output = output * math.sqrt(W[depth - 1].size()[1])
        return output

    def _gradient_loss(self, X, Y, W, THETA):
        """ Return a list of grad, satisfying that W[i] = W[i] - grad[i] for single context x. """
        depth = len(W)
        num_sample = Y.shape[0]
        loss = []
        grad = []
        relu = []
        output = X
        loss.append(output)
        for i in range(0, depth - 1):
            output = torch.mm(W[i], output)
            relu.append(output)
            output = output.clamp(min=0)
            loss.append(output)

        THETA_t = torch.transpose(THETA, 0, 1).view(num_sample, 1, -1)
        output_t = torch.transpose(output, 0, 1).view(num_sample, -1, 1)
        output = torch.bmm(THETA_t, output_t).squeeze().view(1, -1)

        loss.append(output)

        feat = self.feature_extractor(X, W)
        feat_t = torch.transpose(feat, 0, 1).view(num_sample, -1, 1)
        output_t = torch.bmm(THETA_t, feat_t).squeeze().view(1, -1)

        # backward gradient propagation
        back = output_t - Y
        back = back.double()
        grad_t = torch.mm(back, loss[depth - 1].t())
        grad.append(grad_t)

        for i in range(1, depth):
            back = torch.mm(W[depth - i].t(), back)
            back[relu[depth - i - 1] < 0] = 0
            grad_t = torch.mm(back, loss[depth - i - 1].t())
            grad.append(grad_t)

        grad1 = []
        for i in range(0, depth):
            grad1.append(grad[depth - 1 - i] * math.sqrt(W[depth - 1].size()[1]) / len(X[0, :]))

        return grad1

    def _loss(self, X, Y, W, THETA):
        # total loss
        num_sample = len(X[0, :])
        output = self.feature_extractor(X, W)
        THETA_t = torch.transpose(THETA, 0, 1).view(num_sample, 1, -1)
        output_t = torch.transpose(output, 0, 1).view(num_sample, -1, 1)
        output_y = torch.bmm(THETA_t, output_t).squeeze().view(1, -1)

        summ = (Y - output_y).pow(2).sum() / num_sample
        return summ



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from golem.core.optimisers.genetic.operators.operator import PopulationT

T = TypeVar('T')


class AdaptiveParameter(ABC, Generic[T]):
    """Abstract interface for parameters of evolutionary algorithm and its Operators.
    Adaptive parameter is defined by `initial` value and subsequent `next` values.
    Parameter can potentially change on each call to next().
    The specific policy of adaptivity is determined by implementations."""

    @property
    @abstractmethod
    def initial(self) -> T:
        raise NotImplementedError()

    @abstractmethod
    def next(self, population: PopulationT) -> T:
        raise NotImplementedError()


class ConstParameter(AdaptiveParameter[T]):
    """Stub implementation of AdaptiveParameter for constant parameters."""

    def __init__(self, value: T):
        self._value = value

    @property
    def initial(self) -> T:
        return self._value

    def next(self, population: PopulationT) -> T:
        return self._value



import os
from typing import Optional, Union, TYPE_CHECKING

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation

from golem.core.optimisers.genetic.operators.operator import PopulationT
from golem.visualisation.opt_history.history_visualization import HistoryVisualization
from golem.visualisation.opt_history.utils import show_or_save_figure

if TYPE_CHECKING:
    from golem.core.optimisers.opt_history_objects.opt_history import OptHistory


class DiversityLine(HistoryVisualization):
    def visualize(self, show: bool = True, save_path: Optional[Union[os.PathLike, str]] = None):
        """Plots double graphic to estimate diversity of populations during optimization.
        It plots standard deviation of all metrics (y-axis) by generation number (x-axis).
        Additional line show ratio of structurally unique individuals."""
        return plot_diversity_dynamic(self.history, show=show, save_path=save_path, dpi=self.visuals_params['dpi'])


class DiversityPopulation(HistoryVisualization):
    def visualize(self, save_path: Union[os.PathLike, str], fps: int = 4):
        """Creates a GIF with violin-plot estimating distribution of each metric in populations.
        Each frame shows distribution for a particular generation."""
        return plot_diversity_dynamic_gif(self.history, filename=save_path, fps=fps, dpi=self.visuals_params['dpi'])


def compute_fitness_diversity(population: PopulationT) -> np.ndarray:
    """Returns numpy array of standard deviations of fitness values."""
    # substitutes None values
    fitness_values = np.array([ind.fitness.values for ind in population], dtype=float)
    # compute std along each axis while ignoring nan-s
    diversity = np.nanstd(fitness_values, axis=0)
    return diversity


def plot_diversity_dynamic_gif(history: 'OptHistory',
                               filename: Optional[str] = None,
                               fig_size: int = 5,
                               fps: int = 4,
                               dpi: int = 100,
                               ) -> FuncAnimation:
    metric_names = history.objective.metric_names
    # dtype=float removes None, puts np.nan
    # indexed by [population, metric, individual] after transpose (.T)
    pops = history.generations[1:-1]  # ignore initial pop and final choices
    fitness_distrib = [np.array([ind.fitness.values for ind in pop], dtype=float).T
                       for pop in pops]

    # Define bounds on metrics: find min & max on a flattened view of array
    q = 0.025
    lims_max = np.max([np.quantile(pop, 1 - q, axis=1) for pop in fitness_distrib], axis=0)
    lims_min = np.min([np.quantile(pop, q, axis=1) for pop in fitness_distrib], axis=0)

    # Setup the plot
    ncols = max(len(metric_names), 1)
    fig, axs = plt.subplots(ncols=ncols)
    fig.set_size_inches(fig_size * ncols, fig_size)
    axs = np.atleast_1d(np.ravel(axs))

    # Set update function for updating data on the axes
    def update_axes(iframe: int):
        for i, (ax, metric_distrib) in enumerate(zip(axs, fitness_distrib[iframe])):
            # Clear & Prepare axes
            ax: plt.Axes
            ax.clear()
            ax.set_xlim(0.5, 1.5)
            ax.set_ylim(lims_min[i], lims_max[i])
            ax.set_ylabel('Metric value')
            ax.grid()
            # Plot information
            fig.suptitle(f'Population {iframe+1} diversity by metric')
            metric_name = metric_names[i] if metric_names else f"metric{i}"
            ax.set_title(f'{metric_name}, '
                         f'mean={np.mean(metric_distrib).round(3)}, '
                         f'std={np.nanstd(metric_distrib).round(3)}')
            ax.violinplot(metric_distrib,
                          quantiles=[0.25, 0.5, 0.75])

    # Run this function in FuncAnimation
    num_frames = len(fitness_distrib)
    animate = FuncAnimation(
        fig=fig,
        func=update_axes,
        save_count=num_frames,
        interval=200,
    )
    # Save the GIF from animation
    if filename:
        animate.save(filename, fps=fps, dpi=dpi)
    return animate


def plot_diversity_dynamic(history: 'OptHistory',
                           show: bool = True, save_path: Optional[str] = None, dpi: int = 100):
    labels = history.objective.metric_names
    h = history.generations[:-1]  # don't consider final choices
    xs = np.arange(len(h))

    # Compute diversity by metrics
    np_history = np.array([compute_fitness_diversity(pop) for pop in h])
    ys = {label: np_history[:, i] for i, label in enumerate(labels)}
    # Compute number of unique individuals, plot
    ratio_unique = [len(set(ind.graph.descriptive_id for ind in pop)) / len(pop) for pop in h]

    fig, ax = plt.subplots()
    fig.suptitle('Population diversity')
    ax.set_xlabel('Generation')
    ax.set_ylabel('Metrics Std')
    ax.grid()
    line_alpha = 0.8

    for label, metric_std in ys.items():
        ax.plot(xs, metric_std, label=label, alpha=line_alpha)

    ax2 = ax.twinx()
    ax2_color = 'm'  # magenta
    ax2.set_ylabel('Structural uniqueness', color=ax2_color)
    ax2.set_ylim(0.25, 1.05)
    ax2.tick_params(axis='y', labelcolor=ax2_color)
    ax2.plot(xs, ratio_unique, label='unique ratio',
             color=ax2_color, linestyle='dashed', alpha=line_alpha)

    # ask matplotlib for the plotted objects and their labels
    # to put them into single legend for both axes
    lines, labels = ax.get_legend_handles_labels()
    lines2, labels2 = ax2.get_legend_handles_labels()
    ax2.legend(lines + lines2, labels + labels2,
               loc='upper left', bbox_to_anchor=(0., 1.15))

    if show or save_path:
        show_or_save_figure(fig, save_path, dpi)



from golem.core.log import default_log


def warn_requirement(name: str, default_install_path: str, *, should_raise: bool = False):
    """
    :param name: module name failed to load
    :default_install_path: path to requirements than need to be installed
    :param should_raise: bool indicating if ImportError should be raised
    """
    msg = f'"{name}" is not installed, use "pip install {default_install_path}" to fulfil requirement'
    if should_raise:
        raise ImportError(msg)
    else:
        default_log(prefix='Requirements').debug(f'{msg} or ignore this warning')



from typing import Union, Optional, Iterable, List
from golem.core.dag.graph_node import GraphNode
from golem.utilities.data_structures import UniqueList


class LinkedGraphNode(GraphNode):
    """Class for node definition in the directed graph structure
    that directly stores its parent nodes.

    Args:
        nodes_from: parent nodes which information comes from
        content: ``dict`` for the content in the node

    Notes:
        The possible parameters are:
            - ``name`` - name (str) or object that performs actions in this node
            - ``params`` - dictionary with additional information that is used by
                    the object in the ``name`` field (e.g. hyperparameters values)
    """

    def __init__(self, content: Union[dict, str],
                 nodes_from: Optional[Iterable['LinkedGraphNode']] = None):
        # Wrap string into dict if it is necessary
        if isinstance(content, str):
            content = {'name': content}

        self.content: dict = content
        self._nodes_from = UniqueList(nodes_from or ())

        super().__init__()

    @property
    def nodes_from(self) -> List['LinkedGraphNode']:
        return self._nodes_from

    @nodes_from.setter
    def nodes_from(self, nodes: Optional[Iterable['LinkedGraphNode']]):
        self._nodes_from = UniqueList(nodes)

    @property
    def name(self) -> str:
        name = self.content.get('name')
        return str(name) if name is not None else ''

    @property
    def parameters(self) -> dict:
        return self.content.get('params', {})

    @parameters.setter
    def parameters(self, new_parameters):
        if self.content.get('params'):
            self.content['params'].update(new_parameters)
        else:
            self.content['params'] = new_parameters

    def __hash__(self) -> int:
        return hash(self.uid)

    def __str__(self) -> str:
        return str(self.content.get('name', self.uid))

    def __repr__(self) -> str:
        return self.__str__()

    def description(self) -> str:
        label = self.name or self.uid
        # TODO: possibly unify with __repr__ & don't duplicate Operation.description
        if not self.parameters:
            node_label = f'n_{label}'
        elif isinstance(label, str):
            # If there is a string: name of operation (as in json repository)
            node_label = f'n_{label}_{self.parameters}'
        else:
            # If instance of Operation is placed in 'name'
            node_label = label.description(self.parameters)
        return node_label



from datetime import timedelta
from functools import partial
from typing import Type, Optional, Sequence, List

import networkx as nx

from examples.synthetic_graph_evolution.experiment_setup import run_experiments
from examples.synthetic_graph_evolution.generators import postprocess_nx_graph
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.dag.graph import Graph
from golem.core.dag.verification_rules import DEFAULT_DAG_RULES
from golem.core.optimisers.adaptive.operator_agent import MutationAgentTypeEnum
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.base_mutations import MutationTypesEnum
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams, GraphOptimizer, AlgorithmParameters
from golem.metrics.graph_metrics import spectral_dist, size_diff, degree_distance


def graph_search_setup(target_graph: Optional[nx.DiGraph] = None,
                       objective: Optional[Objective] = None,
                       optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer,
                       algorithm_parameters: Optional[AlgorithmParameters] = None,
                       node_types: Sequence[str] = ('x',),
                       timeout: Optional[timedelta] = None,
                       num_iterations: Optional[int] = None,
                       initial_graph_sizes: Optional[List[int]] = None,
                       initial_graphs: List[Graph] = None,
                       pop_size: int = None):
    if target_graph is not None and objective is not None:
        raise ValueError('Please provide either target or objective, not both')
    elif target_graph is not None:
        # Setup objective that measures some graph-theoretic similarity measure
        objective = Objective(
            quality_metrics={
                'sp_adj': partial(spectral_dist, target_graph, kind='adjacency'),
                'sp_lapl': partial(spectral_dist, target_graph, kind='laplacian'),
            },
            complexity_metrics={
                'graph_size': partial(size_diff, target_graph),
                'degree': partial(degree_distance, target_graph),
            },
            is_multi_objective=True
        )
        max_graph_size = target_graph.number_of_nodes()
    elif objective is not None:
        max_graph_size = 1000
    else:
        raise ValueError()

    # Setup parameters
    requirements = GraphRequirements(
        max_arity=max_graph_size,
        max_depth=max_graph_size,
        early_stopping_timeout=10,
        early_stopping_iterations=num_iterations // 3 if num_iterations else None,
        keep_n_best=4,
        timeout=timeout,
        num_of_generations=num_iterations,
        n_jobs=-1,
        history_dir=None,
    )
    default_gp_params = GPAlgorithmParameters(
        adaptive_mutation_type=MutationAgentTypeEnum.random,
        pop_size=pop_size or 21,
        multi_objective=objective.is_multi_objective,
        genetic_scheme_type=GeneticSchemeTypesEnum.generational,
        mutation_types=[
            MutationTypesEnum.single_add,
            MutationTypesEnum.single_edge,
            MutationTypesEnum.single_drop
        ],
        crossover_types=[CrossoverTypesEnum.none]
    )
    gp_params = algorithm_parameters or default_gp_params
    graph_gen_params = GraphGenerationParams(
        adapter=BaseNetworkxAdapter(),
        rules_for_constraint=DEFAULT_DAG_RULES,
        available_node_types=node_types,
    )

    # Generate simple initial population with line graphs
    if not initial_graphs:
        if not initial_graph_sizes:
            initial_graph_sizes = [7] * gp_params.pop_size
        initial_graphs = [postprocess_nx_graph(nx.random_tree(initial_graph_sizes[i], create_using=nx.DiGraph),
                                               node_labels=node_types)
                          for i in range(gp_params.pop_size)]
    # Build the optimizer
    optimiser = optimizer_cls(objective, initial_graphs, requirements, graph_gen_params, gp_params)
    return optimiser, objective


if __name__ == '__main__':
    results_log = run_experiments(optimizer_setup=graph_search_setup,
                                  optimizer_cls=EvoGraphOptimizer,
                                  graph_names=['gnp'],
                                  graph_sizes=[50],
                                  num_trials=1,
                                  trial_timeout=15,
                                  trial_iterations=1000,
                                  visualize=True)
    print(results_log)



from typing import Optional

import numpy as np

from golem.core.constants import MIN_POP_SIZE, EVALUATION_ATTEMPTS_NUMBER
from golem.core.log import default_log
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.crossover import Crossover
from golem.core.optimisers.genetic.operators.mutation import Mutation
from golem.core.optimisers.genetic.operators.operator import PopulationT, EvaluationOperator
from golem.core.optimisers.genetic.operators.selection import Selection
from golem.core.optimisers.populational_optimizer import EvaluationAttemptsError
from golem.utilities.data_structures import ensure_wrapped_in_sequence


class ReproductionController:
    """
    Task of the Reproduction Controller is to reproduce population
    while keeping population size as specified in optimizer settings.

    It implements a simple proportional controller that compensates for
    invalid results each generation by computing average ratio of valid results.
    Invalid results include cases when Operators, Evaluator or GraphVerifier
    return output population that's smaller than the input population.

    Example.
    Let's say we need a population of size 50. Let's say about 20% of individuals
    are *usually* evaluated with an error. If we take select only 50 for the new population,
    we will get about 40 valid ones. Not enough. Therefore, we need to take more.
    How much more? Approximately by `target_pop_size / mean_success_rate = 50 / 0.8 ~= 62'.
    Here `mean_success_rate` estimates number of successfully evaluated individuals.
    Then we request 62, then approximately 62*0.8~=50 of them are valid in the end,
    and we achieve target size more reliably. This runs in a loop to control stochasticity.

    Args:
        parameters: genetic algorithm parameters.
        selection: operator used in reproduction.
        mutation: operator used in reproduction.
        crossover: operator used in reproduction.
        window_size: size in iterations of the moving window to compute reproduction success rate.
    """

    def __init__(self,
                 parameters: GPAlgorithmParameters,
                 selection: Selection,
                 mutation: Mutation,
                 crossover: Crossover,
                 window_size: int = 10,
                 ):
        self.parameters = parameters
        self.selection = selection
        self.mutation = mutation
        self.crossover = crossover

        self._minimum_valid_ratio = parameters.required_valid_ratio * 0.5
        self._window_size = window_size
        self._success_rate_window = np.full(self._window_size, 1.0)

        self._log = default_log(self)

    @property
    def mean_success_rate(self) -> float:
        """Returns mean success rate of reproduction + evaluation,
        fraction of how many individuals were reproduced and mutated successfully.
        Computed as average fraction for the last N iterations (N = window size param)"""
        return float(np.mean(self._success_rate_window))

    def reproduce_uncontrolled(self,
                               population: PopulationT,
                               evaluator: EvaluationOperator,
                               pop_size: Optional[int] = None,
                               ) -> PopulationT:
        """Reproduces and evaluates population (select, crossover, mutate).
        Doesn't implement any additional checks on population.
        """
        # If operators can return unchanged individuals from previous population
        # (e.g. both Mutation & Crossover are not applied with some probability)
        # then there's a probability that duplicate individuals can appear

        # TODO: it can't choose more than len(population)!
        #  It can be faster if it could.
        selected_individuals = self.selection(population, pop_size)
        new_population = self.crossover(selected_individuals)
        new_population = ensure_wrapped_in_sequence(self.mutation(new_population))
        new_population = evaluator(new_population)
        return new_population

    def reproduce(self,
                  population: PopulationT,
                  evaluator: EvaluationOperator
                  ) -> PopulationT:
        """Reproduces and evaluates population (select, crossover, mutate).
        Implements additional checks on population to ensure that population size
        follows required population size.
        """
        total_target_size = self.parameters.pop_size  # next population size
        collected_next_population = {}
        for i in range(EVALUATION_ATTEMPTS_NUMBER):
            # Estimate how many individuals we need to complete new population
            # based on average success rate of valid results
            residual_size = total_target_size - len(collected_next_population)
            residual_size = max(MIN_POP_SIZE,
                                int(residual_size / self.mean_success_rate))
            residual_size = min(len(population), residual_size)

            # Reproduce the required number of individuals that equals residual size
            partial_next_population = self.reproduce_uncontrolled(population, evaluator, residual_size)
            # Avoid duplicate individuals that can come unchanged from previous population
            collected_next_population.update({ind.uid: ind for ind in partial_next_population})

            # Keep running average of transform success rate (if sample is big enough)
            if len(partial_next_population) >= MIN_POP_SIZE:
                valid_ratio = len(partial_next_population) / residual_size
                self._success_rate_window = np.roll(self._success_rate_window, shift=1)
                self._success_rate_window[0] = valid_ratio

            # Successful return: got enough individuals
            if len(collected_next_population) >= total_target_size * self.parameters.required_valid_ratio:
                self._log.info(f'Reproduction achieved pop size {len(collected_next_population)}'
                               f' using {i+1} attempt(s) with success rate {self.mean_success_rate:.3f}')
                return list(collected_next_population.values())[:total_target_size]
        else:
            # If number of evaluation attempts is exceeded return a warning or raise exception
            helpful_msg = ('Check objective, constraints and evo operators. '
                           'Possibly they return too few valid individuals.')

            if len(collected_next_population) >= total_target_size * self._minimum_valid_ratio:
                self._log.warning(f'Could not achieve required population size: '
                                  f'have {len(collected_next_population)},'
                                  f' required {total_target_size}!\n' + helpful_msg)
                return list(collected_next_population.values())
            else:
                raise EvaluationAttemptsError('Could not collect valid individuals'
                                              ' for next population.' + helpful_msg)



from typing import Tuple, List

from golem.core.dag.graph_node import GraphNode


class Edge:
    def __init__(self, parent_node: GraphNode, child_node: GraphNode):
        self.parent_node = parent_node
        self.child_node = child_node
        self.rating = None

    @staticmethod
    def from_tuple(edges_in_tuple: List[Tuple[GraphNode, GraphNode]]) -> List['Edge']:
        edges = []
        for edge in edges_in_tuple:
            edges.append(Edge(child_node=edge[1], parent_node=edge[0]))
        return edges



import json
import os
from importlib import import_module
from inspect import isclass, isfunction, ismethod, signature
from json import JSONDecoder, JSONEncoder
from typing import Any, Callable, Dict, Optional, Type, TypeVar, Union

from golem.core.dag.linked_graph import LinkedGraph
from golem.core.dag.linked_graph_node import LinkedGraphNode
from golem.core.log import default_log

# NB: at the end of module init happens registration of default class coders

INSTANCE_OR_CALLABLE = TypeVar('INSTANCE_OR_CALLABLE', object, Callable)
EncodeCallable = Callable[[INSTANCE_OR_CALLABLE], Dict[str, Any]]
DecodeCallable = Callable[[Type[INSTANCE_OR_CALLABLE], Dict[str, Any]], INSTANCE_OR_CALLABLE]

MODULE_X_NAME_DELIMITER = '/'
CLASS_PATH_KEY = '_class_path'

# Mapping between class paths for backward compatibility for renamed/moved classes
LEGACY_CLASS_PATHS = {
    'fedot.core.optimisers.gp_comp.individual/Individual':
        'golem.core.optimisers.opt_history_objects.individual/Individual',
    'fedot.core.optimisers.gp_comp.individual/ParentOperator':
        'golem.core.optimisers.opt_history_objects.parent_operator/ParentOperator',
    'fedot.core.optimisers.opt_history/OptHistory':
        'golem.core.optimisers.opt_history_objects.opt_history/OptHistory',

    'fedot.core.dag.graph_node/GraphNode':
        'golem.core.dag.linked_graph_node/LinkedGraphNode',
    'fedot.core.dag.graph_operator/GraphOperator':
        'golem.core.dag.linked_graph/LinkedGraph',
    'fedot.core.dag.graph_operator/GraphOperator._empty_postprocess':
        'golem.core.dag.linked_graph/LinkedGraph._empty_postprocess',

    # for fedot->golem transition
    'fedot.core.optimisers.objective.objective/Objective':
        'golem.core.optimisers.objective.objective/Objective',
    'fedot.core.optimisers.fitness.fitness/Fitness':
        'golem.core.optimisers.fitness.fitness/Fitness',
}

# for fedot->golem transition
# NB: must be ordered from specific to top-level modules
LEGACY_MODULE_PATHS = {
    'fedot.core.optimisers.gp_comp.individual': 'golem.core.optimisers.opt_history_objects.individual',
    'fedot.core.optimisers.opt_history_objects': 'golem.core.optimisers.opt_history_objects',
    'fedot.core.optimisers.gp_comp': 'golem.core.optimisers.genetic',
    'fedot.core.optimisers.graph': 'golem.core.optimisers.graph',
    'fedot.core.optimisers.objective.objective': 'golem.core.optimisers.objective.objective',
    'fedot.core.optimisers.fitness': 'golem.core.optimisers.fitness',

    'fedot.core.log': 'golem.core.log',
    'fedot.core.adapter': 'golem.core.adapter',
    'fedot.core.dag': 'golem.core.dag',
    'fedot.core.utilities': 'golem.core.utilities',
}


class Serializer(JSONEncoder, JSONDecoder):
    _to_json = 'to_json'
    _from_json = 'from_json'

    CODERS_BY_TYPE = {}

    __default_coders_initialized = False

    def __init__(self, *args, **kwargs):
        for base_class, coder_name in [(JSONEncoder, 'default'), (JSONDecoder, 'object_hook')]:
            base_kwargs = {k: kwargs[k] for k in kwargs.keys() & signature(base_class.__init__).parameters}
            base_kwargs[coder_name] = getattr(self, coder_name)
            base_class.__init__(self, **base_kwargs)

        if not self.__default_coders_initialized:
            Serializer._register_default_coders()
            self.__default_coders_initialized = True

    @staticmethod
    def _register_default_coders():
        from uuid import UUID

        from golem.core.dag.graph import Graph
        from golem.core.dag.linked_graph_node import LinkedGraphNode
        from golem.core.optimisers.opt_history_objects.individual import Individual
        from golem.core.optimisers.opt_history_objects.generation import Generation
        from golem.core.optimisers.opt_history_objects.opt_history import OptHistory
        from golem.core.optimisers.opt_history_objects.parent_operator import ParentOperator
        from golem.core.optimisers.fitness.fitness import Fitness
        from golem.core.optimisers.objective.objective import ObjectiveInfo
        from golem.utilities.data_structures import ComparableEnum

        from .any_serialization import any_from_json, any_to_json

        from .coders import (
            enum_from_json,
            enum_to_json,
            graph_from_json,
            graph_node_to_json,
            opt_history_from_json,
            opt_history_to_json,
            parent_operator_from_json,
            parent_operator_to_json,
            uuid_from_json,
            uuid_to_json
        )

        _to_json = Serializer._to_json
        _from_json = Serializer._from_json
        basic_serialization = {_to_json: any_to_json, _from_json: any_from_json}

        default_coders = {
            Fitness: basic_serialization,
            ObjectiveInfo: basic_serialization,
            Individual: basic_serialization,
            Generation: basic_serialization,
            LinkedGraphNode: {_to_json: graph_node_to_json, _from_json: any_from_json},
            Graph: {_to_json: any_to_json, _from_json: graph_from_json},
            OptHistory: {_to_json: opt_history_to_json, _from_json: opt_history_from_json},
            ParentOperator: {_to_json: parent_operator_to_json, _from_json: parent_operator_from_json},
            UUID: {_to_json: uuid_to_json, _from_json: uuid_from_json},
            ComparableEnum: {_to_json: enum_to_json, _from_json: enum_from_json},
        }
        Serializer.CODERS_BY_TYPE.update(default_coders)

    @staticmethod
    def register_coders(cls: Type[INSTANCE_OR_CALLABLE],
                        to_json: Optional[EncodeCallable[INSTANCE_OR_CALLABLE]] = None,
                        from_json: Optional[DecodeCallable[INSTANCE_OR_CALLABLE]] = None,
                        overwrite: bool = False,
                        ) -> Type[INSTANCE_OR_CALLABLE]:
        """Registers classes as json-serializable so that they can be used by `Serializer`.

        Supports 3 alternative usages:

        - Default serialization is used (functions `any_to_json` and `any_from_json`).
        - Custom serialization functions `to_json` & `from_json` are provided explicitly as arguments.
        - Custom serialization functions `to_json` & `from_json` are defined in the class.

        Args:
            cls: registered class
            to_json: custom encoding function that returns json Dict.
             Optional, if None then default is used.
            from_json: custom decoding function that returns class given a Dict.
             Optional, if None then default is used.
            overwrite: flag that allows to overwrite existing registered coders.
             False by default: method raises AttributeError in attempt to overwrite coders.

        Returns:
            cls: class that is registered in serializer
        """
        if cls is None:
            raise TypeError('Class must not be None.')
        from .any_serialization import any_from_json, any_to_json

        # get provided coders or coders defined in the class itself or default universal coders
        coders = {Serializer._to_json: to_json or getattr(cls, Serializer._to_json, any_to_json),
                  Serializer._from_json: from_json or getattr(cls, Serializer._from_json, any_from_json)}

        if cls not in Serializer.CODERS_BY_TYPE or overwrite:
            Serializer.CODERS_BY_TYPE[cls] = coders
        else:
            raise AttributeError(f'Object {cls} already has serializer coders registered.')

        return cls

    @staticmethod
    def _get_field_checker(obj: Union[INSTANCE_OR_CALLABLE, Type[INSTANCE_OR_CALLABLE]]) -> Callable[..., bool]:
        if isclass(obj):
            return issubclass
        return isinstance

    @staticmethod
    def _get_base_type(obj: Union[INSTANCE_OR_CALLABLE, Type[INSTANCE_OR_CALLABLE]]) -> Optional[Type]:
        contains = Serializer._get_field_checker(obj)
        for k_type in Serializer.CODERS_BY_TYPE:
            if contains(obj, k_type):
                return k_type
        return None

    @staticmethod
    def _get_coder_by_type(coder_type: Type, coder_aim: str):
        return Serializer.CODERS_BY_TYPE[coder_type][coder_aim]

    @staticmethod
    def dump_path_to_obj(obj: INSTANCE_OR_CALLABLE) -> Dict[str, str]:
        """
        Dumps the full path (module + name) to the input object into the dict

        :param obj: object which path should be resolved (class, function or method)

        :return: dict[str, str] with path to the object
        """
        if isclass(obj) or isfunction(obj) or ismethod(obj):
            obj_name = obj.__qualname__
        else:
            obj_name = obj.__class__.__qualname__

        if getattr(obj, '__module__', None) is not None:
            obj_module = obj.__module__
        else:
            obj_module = obj.__class__.__module__
        return {
            CLASS_PATH_KEY: f'{obj_module}{MODULE_X_NAME_DELIMITER}{obj_name}'
        }

    def default(self, obj: INSTANCE_OR_CALLABLE) -> Dict[str, Any]:
        """
        Tries to encode objects that are not simply json-encodable to JSON-object

        :param obj: object to be encoded (class, function or method)

        :return: dict[str, Any] which is in fact json object
        """
        if isfunction(obj) or ismethod(obj):
            return Serializer.dump_path_to_obj(obj)
        base_type = Serializer._get_base_type(obj)
        if base_type is not None:
            encoded = Serializer._get_coder_by_type(base_type, Serializer._to_json)(obj)
            if CLASS_PATH_KEY not in encoded:
                encoded.update(Serializer.dump_path_to_obj(obj))
            return encoded

        return JSONEncoder.default(self, obj)

    @staticmethod
    def _get_class(json_obj: dict) -> Optional[Type[INSTANCE_OR_CALLABLE]]:
        """
        Gets the object type from the class_path

        :param class_path: full path (module + name) of the class

        :return: class, function or method type
        """
        class_path = json_obj[CLASS_PATH_KEY]
        class_path = LEGACY_CLASS_PATHS.get(class_path, class_path)
        module_name, class_name = class_path.split(MODULE_X_NAME_DELIMITER)
        module_name = Serializer._legacy_module_map(module_name)

        try:
            obj_cls = import_module(module_name)
        except ImportError as ex:
            obj_cls = Serializer._import_as_base_class(json_obj)
            if not obj_cls:
                default_log('Serializer').info(
                    f'Object was not decoded and will be stored as a dict '
                    f'because of an ImportError: {ex}.')
            else:
                default_log('Serializer').info(
                    f'Object was decoded as {obj_cls} and not as an original class '
                    f'because of an ImportError: {ex}.')
            return obj_cls

        for sub in class_name.split('.'):
            obj_cls = getattr(obj_cls, sub)
        return obj_cls

    @staticmethod
    def _legacy_module_map(module_path: str) -> str:
        for legacy_prefix, new_prefix in LEGACY_MODULE_PATHS.items():
            if module_path.startswith(legacy_prefix):
                return module_path.replace(legacy_prefix, new_prefix)
        return module_path

    @staticmethod
    def _is_bound_method(method: Callable) -> bool:
        return hasattr(method, '__self__')

    @staticmethod
    def _import_as_base_class(json_obj: dict) \
            -> Optional[Union[Type[LinkedGraph], Type[LinkedGraphNode]]]:
        linked_graph_keys = {'_nodes', '_postprocess_nodes'}
        linked_node_keys = {'content', '_nodes_from', 'uid'}
        if linked_graph_keys.issubset(json_obj.keys()):
            return LinkedGraph
        elif linked_node_keys.issubset(json_obj.keys()):
            return LinkedGraphNode
        else:
            return None

    @staticmethod
    def object_hook(json_obj: Dict[str, Any]) -> Union[INSTANCE_OR_CALLABLE, dict]:
        """
        Decodes every JSON-object to python class/func object or just returns dict

        :param json_obj: dict[str, Any] to be decoded into Python class, function or
            method object only if it has some special fields

        :return: Python class, function or method object OR input if it's just a regular dict
        """
        if CLASS_PATH_KEY in json_obj:
            obj_cls = Serializer._get_class(json_obj)
            del json_obj[CLASS_PATH_KEY]
            base_type = Serializer._get_base_type(obj_cls)
            if isclass(obj_cls) and base_type is not None:
                coder = Serializer._get_coder_by_type(base_type, Serializer._from_json)
                # call with a right num of arguments
                if Serializer._is_bound_method(coder):
                    return coder(json_obj)
                else:
                    return coder(obj_cls, json_obj)
            elif isfunction(obj_cls) or ismethod(obj_cls):
                return obj_cls
            else:
                return json_obj
        return json_obj


def default_save(obj: Any, json_file_path: Optional[Union[str, os.PathLike]] = None) -> Optional[str]:
    """ Default save to json using Serializer """
    if json_file_path is None:
        return json.dumps(obj, indent=4, cls=Serializer)
    with open(json_file_path, mode='w') as json_file:
        json.dump(obj, json_file, indent=4, cls=Serializer)
        return None


def default_load(json_str_or_file_path: Union[str, os.PathLike]) -> Any:
    """ Default load from json using Serializer """

    def load_as_file_path():
        with open(json_str_or_file_path, mode='r') as json_file:
            return json.load(json_file, cls=Serializer)

    def load_as_json_str():
        return json.loads(json_str_or_file_path, cls=Serializer)

    if isinstance(json_str_or_file_path, os.PathLike):
        return load_as_file_path()

    try:
        return load_as_json_str()
    except json.JSONDecodeError:
        return load_as_file_path()


def register_serializable(cls: Optional[Type[INSTANCE_OR_CALLABLE]] = None,
                          to_json: Optional[EncodeCallable] = None,
                          from_json: Optional[DecodeCallable] = None,
                          overwrite: bool = False,
                          add_save_load: bool = False,
                          ) -> Type[INSTANCE_OR_CALLABLE]:
    """Decorator for registering classes as json-serializable.
    Relies on :py:class:`golem.serializers.serializer.Serializer.register_coders`.
    Optionally adds `save` and `load` methods to the class for json (de)serialization.

    Args:
        cls: decorated class
        to_json: custom encoding function that returns json Dict.
         Optional, if None then default is used.
        from_json: custom decoding function that returns class given a Dict.
         Optional, if None then default is used.
        overwrite: flag that allows to overwrite existing registered coders.
         False by default: method raises AttributeError in attempt to overwrite coders.
        add_save_load: if True, then `save` and `load` methods are added to the class
         that (de)serialize the class (from)to the file using `Serializer`.
         See `default_save` & `default_load` functions.

    Returns:
        cls: class that is registered in serializer
    """
    _save = 'save'
    _load = 'load'

    def make_serializable(cls):
        Serializer.register_coders(cls, to_json, from_json, overwrite)
        if add_save_load:
            if hasattr(cls, _save) or hasattr(cls, _load):
                raise ValueError(f'Class {cls} already have `save` and/or `load` methods, can not overwrite them.')
            setattr(cls, _save, default_save)
            setattr(cls, _load, default_load)
        return cls

    # See if we're being called as @decorator() (with parens).
    if cls is None:
        # We're called with parens.
        return make_serializable

    # We're called as `@decorator` without parens.
    return make_serializable(cls)



"""
********
Matrices
********

Matrices associated with graphs. Also contains linear algebraic helper functions.
"""

from scipy import sparse as sps
from scipy.sparse import issparse
import numpy as np

_eps = 10 ** (-10)  # a small parameter


######################
## Helper Functions ##
######################


def _flat(D):
    """Flatten column or row matrices, as well as arrays."""
    if issparse(D):
        raise ValueError('Cannot flatten sparse matrix.')
    d_flat = np.array(D).flatten()
    return d_flat


def _pad(A, N):
    """Pad A so A.shape is (N,N)"""
    n, _ = A.shape
    if n >= N:
        return A
    else:
        if issparse(A):
            # thrown if we try to np.concatenate sparse matrices
            side = sps.csr_matrix((n, N - n))
            bottom = sps.csr_matrix((N - n, N))
            A_pad = sps.hstack([A, side])
            A_pad = sps.vstack([A_pad, bottom])
        else:
            side = np.zeros((n, N - n))
            bottom = np.zeros((N - n, N))
            A_pad = np.concatenate([A, side], axis=1)
            A_pad = np.concatenate([A_pad, bottom])
        return A_pad


########################
## Matrices of Graphs ##
########################


def degree_matrix(A):
    """Diagonal degree matrix of graph with adjacency matrix A

    Parameters
    ----------
    A : matrix
        Adjacency matrix

    Returns
    -------
    D : SciPy sparse matrix
        Diagonal matrix of degrees.
    """
    n, m = A.shape
    degs = _flat(A.sum(axis=1))
    D = sps.spdiags(degs, [0], n, n, format='csr')
    return D


def laplacian_matrix(A, normalized=False):
    """Diagonal degree matrix of graph with adjacency matrix A

    Parameters
    ----------
    A : matrix
        Adjacency matrix
    normalized : Bool, optional (default=False)
        If true, then normalized laplacian is returned.

    Returns
    -------
    L : SciPy sparse matrix
        Combinatorial laplacian matrix.
    """
    n, m = A.shape
    D = degree_matrix(A)
    L = D - A
    if normalized:
        degs = _flat(A.sum(axis=1))
        rootD = sps.spdiags(np.power(degs, -1 / 2), [0], n, n, format='csr')
        L = rootD * L * rootD
    return L



import numpy as np

MAX_GRAPH_GEN_ATTEMPTS = 1000
MAX_TUNING_METRIC_VALUE = np.inf
MIN_TIME_FOR_TUNING_IN_SEC = 3
# Max number of evaluations attempts to collect the next pop; See usages.
EVALUATION_ATTEMPTS_NUMBER = 5
# Min pop size to avoid getting stuck in local maximum during optimization.
MIN_POP_SIZE = 5



from random import choice
from typing import Optional, Sequence

from golem.core.dag.graph import Graph
from golem.core.optimisers.archive import GenerationKeeper
from golem.core.optimisers.genetic.evaluation import SequentialDispatcher
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.operator import EvaluationOperator
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.objective import Objective, ObjectiveFunction
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphOptimizer, GraphGenerationParams
from golem.core.optimisers.timer import OptimisationTimer
from golem.utilities.grouped_condition import GroupedCondition


class RandomSearchOptimizer(GraphOptimizer):

    def __init__(self,
                 objective: Objective,
                 initial_graphs: Sequence[Graph] = None,
                 requirements: Optional[GraphRequirements] = None,
                 graph_generation_params: Optional[GraphGenerationParams] = None,
                 graph_optimizer_params: Optional[GPAlgorithmParameters] = None):
        super().__init__(objective, initial_graphs, requirements, graph_generation_params, graph_optimizer_params)
        self.timer = OptimisationTimer(timeout=self.requirements.timeout)
        self.generations = GenerationKeeper(self.objective, keep_n_best=requirements.keep_n_best)
        self.current_iteration_num = 0
        self.best_individual = None
        self.stop_optimization = \
            GroupedCondition(results_as_message=True).add_condition(
                lambda: self.timer.is_time_limit_reached(self.current_iteration_num),
                'Optimisation stopped: Time limit is reached'
            ).add_condition(
                lambda: requirements.num_of_generations is not None and
                self.current_iteration_num >= requirements.num_of_generations,
                'Optimisation stopped: Max number of iterations reached')

    def optimise(self, objective: ObjectiveFunction) -> Sequence[OptGraph]:

        dispatcher = SequentialDispatcher(self.graph_generation_params.adapter)
        evaluator = dispatcher.dispatch(objective, self.timer)

        self.current_iteration_num = 0

        with self.timer, self._progressbar as pbar:
            self.best_individual = self._eval_initial_individual(evaluator)
            self._update_best_individual(self.best_individual, 'initial_assumptions')
            while not self.stop_optimization():
                new_individual = self._generate_new_individual()
                evaluator([new_individual])
                self.current_iteration_num += 1
                self._update_best_individual(new_individual)
                pbar.update()
        self._update_best_individual(self.best_individual, 'final_choices')
        pbar.close()
        return [self.best_individual.graph]

    def _update_best_individual(self, new_individual: Individual, label: Optional[str] = None):
        if new_individual.fitness >= self.best_individual.fitness:
            self.best_individual = new_individual

        self.generations.append([new_individual])

        self.log.info(f'Spent time: {round(self.timer.minutes_from_start, 1)} min')
        self.log.info(f'Iteration num {self.current_iteration_num}: '
                      f'Best individuals fitness {str(self.generations)}')

        self.history.add_to_history([new_individual], label)
        self.history.add_to_archive_history(self.generations.best_individuals)

    def _eval_initial_individual(self, evaluator: EvaluationOperator) -> Individual:
        init_ind = Individual(choice(self.initial_graphs)) if self.initial_graphs else self._generate_new_individual()
        evaluator([init_ind])
        return init_ind

    def _generate_new_individual(self) -> Individual:
        new_graph = self.graph_generation_params.random_graph_factory(self.requirements)
        return Individual(new_graph)



from abc import abstractmethod
from typing import Any

import numpy as np

from golem.core.dag.graph import Graph


class SurrogateModel:
    """
    Model for evaluating fitness function without time-consuming fitting pipeline
    """
    @abstractmethod
    def __call__(self, graph: Graph, **kwargs: Any):
        raise NotImplementedError()


class RandomValuesSurrogateModel(SurrogateModel):
    """
        Model for evaluating fitness function based on returning random values for any model
    """
    def __call__(self, graph: Graph, **kwargs: Any):
        return np.random.random(1)



from typing import List

from golem.structural_analysis.graph_sa.results.base_sa_approach_result import BaseSAApproachResult
from golem.structural_analysis.graph_sa.results.deletion_sa_approach_result import \
    DeletionSAApproachResult
from golem.structural_analysis.graph_sa.results.replace_sa_approach_result import \
    ReplaceSAApproachResult
from golem.structural_analysis.graph_sa.results.utils import EntityTypesEnum

NODE_DELETION = 'NodeDeletionAnalyze'
NODE_REPLACEMENT = 'NodeReplaceOperationAnalyze'
SUBTREE_DELETION = 'SubtreeDeletionAnalyze'
EDGE_DELETION = 'EdgeDeletionAnalyze'
EDGE_REPLACEMENT = 'EdgeReplaceOperationAnalyze'


class StructuralAnalysisResultsRepository:
    approaches_dict = {NODE_DELETION: {'result_class': DeletionSAApproachResult},
                       NODE_REPLACEMENT: {'result_class': ReplaceSAApproachResult},
                       SUBTREE_DELETION: {'result_class': DeletionSAApproachResult},
                       EDGE_DELETION: {'result_class': DeletionSAApproachResult},
                       EDGE_REPLACEMENT: {'result_class': ReplaceSAApproachResult}}

    def get_method_by_result_class(self, result_class: BaseSAApproachResult, entity_class: str) -> str:
        for method in self.approaches_dict.keys():
            if self.approaches_dict[method]['result_class'] == result_class.__class__ \
                    and entity_class in method.lower():
                return method

    def get_class_by_str(self, result_str: str) -> BaseSAApproachResult:
        for method in self.approaches_dict.keys():
            if result_str == method:
                return self.approaches_dict[method]['result_class']


class ObjectSAResult:
    """ Class specifying results of Structural Analysis for one entity(node or edge). """

    def __init__(self, entity_idx: str, entity_type: EntityTypesEnum):
        self.entity_idx = entity_idx
        self.entity_type = entity_type
        self.result_approaches: List[BaseSAApproachResult] = []

    def get_worst_result(self, metric_idx_to_optimize_by: int) -> float:
        """ Returns the worst result among all result classes. """
        worst_results = []
        for approach in self.result_approaches:
            worst_results.append(approach.get_worst_result(metric_idx_to_optimize_by=metric_idx_to_optimize_by))
        if not worst_results:
            return 0
        return max(worst_results)

    def get_worst_result_with_names(self, metric_idx_to_optimize_by: int) -> dict:
        """ Returns worst result with additional information. """
        worst_result = self.get_worst_result(metric_idx_to_optimize_by=metric_idx_to_optimize_by)
        for approach in self.result_approaches:
            if approach.get_worst_result(metric_idx_to_optimize_by=metric_idx_to_optimize_by) == worst_result:
                sa_approach_name = self._get_approach_name(approach=approach)
                result = {'entity_idx': self.entity_idx, 'approach_name': sa_approach_name}
                result.update(approach.get_worst_result_with_names(metric_idx_to_optimize_by=metric_idx_to_optimize_by))
                return result

    def add_result(self, result: BaseSAApproachResult):
        self.result_approaches.append(result)

    def get_dict_results(self) -> dict:
        """ Returns dict representation of results. """
        results = dict()
        for approach in self.result_approaches:
            sa_approach_name = self._get_approach_name(approach=approach)
            results[sa_approach_name] = approach.get_dict_results()
        return {self.entity_idx: results}

    def _get_approach_name(self, approach: BaseSAApproachResult) -> str:
        sa_approach_name = StructuralAnalysisResultsRepository() \
            .get_method_by_result_class(approach, self.entity_type.value)
        return sa_approach_name



from pathlib import Path
from typing import List

import setuptools

# The directory containing this file
HERE = Path(__file__).parent.resolve()

# The text of the README file
NAME = 'thegolem'
VERSION = '0.4.0'
AUTHOR = 'NSS Lab'
SHORT_DESCRIPTION = 'Framework for Graph Optimization and Learning by Evolutionary Methods'

README = Path(HERE, 'README_en.rst').read_text(encoding='utf-8')
URL = 'https://github.com/aimclub/GOLEM'
REQUIRES_PYTHON = '>=3.8'
LICENSE = 'BSD 3-Clause'


def _readlines(*names: str, **kwargs) -> List[str]:
    encoding = kwargs.get('encoding', 'utf-8')
    lines = Path(__file__).parent.joinpath(*names).read_text(encoding=encoding).splitlines()
    return list(map(str.strip, lines))


def _extract_requirements(file_name: str):
    return [line for line in _readlines(file_name) if line and not line.startswith('#')]


def _get_requirements(req_name: str):
    requirements = _extract_requirements(req_name)
    return requirements


setuptools.setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email='itmo.nss.team@gmail.com',
    description=SHORT_DESCRIPTION,
    long_description=README,
    long_description_content_type='text/x-rst',
    url=URL,
    python_requires=REQUIRES_PYTHON,
    license=LICENSE,
    packages=setuptools.find_packages(exclude=['test*']),
    include_package_data=True,
    install_requires=_get_requirements('requirements.txt'),
    extras_require={
        key: _get_requirements(Path('other_requirements', f'{key}.txt'))
        for key in ('docs', 'profilers', 'molecules', 'adaptive')
    },
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ],
)



from joblib import cpu_count


def determine_n_jobs(n_jobs=-1, logger=None):
    cpu_num = cpu_count()
    if n_jobs > cpu_num:
        n_jobs = cpu_num
    elif n_jobs <= 0:
        if n_jobs <= -cpu_num - 1 or n_jobs == 0:
            raise ValueError(f"Unproper `n_jobs` = {n_jobs}. "
                             f"`n_jobs` should be between ({-cpu_num}, {cpu_num}) except 0")
        n_jobs = cpu_num + 1 + n_jobs
    if logger:
        logger.info(f"Number of used CPU's: {n_jobs}")
    return n_jobs



import os.path
import random
from datetime import timedelta
from functools import partial
from pprint import pprint

import pandas as pd
import seaborn as sns

from typing import List, Callable, Sequence, Optional, Dict, Tuple

from matplotlib import pyplot as plt
from sklearn.cluster import MiniBatchKMeans

from examples.adaptive_optimizer.mab_experiment_different_targets import get_graph_gp_params
from examples.adaptive_optimizer.utils import plot_action_values
from examples.synthetic_graph_evolution.generators import generate_labeled_graph
from examples.synthetic_graph_evolution.graph_search import graph_search_setup
from examples.synthetic_graph_evolution.utils import draw_graphs_subplots
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.dag.graph import Graph
from golem.core.optimisers.adaptive.context_agents import ContextAgentTypeEnum
from golem.core.optimisers.adaptive.operator_agent import MutationAgentTypeEnum

from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.operators.base_mutations import MutationTypesEnum
from golem.core.optimisers.genetic.operators.operator import PopulationT
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.optimizer import GraphOptimizer
from golem.core.paths import project_root
from golem.visualisation.opt_history.fitness_line import MultipleFitnessLines


class MABSyntheticExperimentHelper:
    """ Class to provide synthetic experiments without data to compare MABs. """

    def __init__(self, launch_num: int, timeout: float, bandits_to_compare: List[MutationAgentTypeEnum],
                 context_agent_types: List[ContextAgentTypeEnum], bandit_labels: List[str] = None,
                 path_to_save: str = None, is_visualize: bool = False, n_clusters: Optional[int] = None,
                 decaying_factors: List[float] = None, window_sizes: List[int] = None):
        self.launch_num = launch_num
        self.timeout = timeout
        self.bandits_to_compare = bandits_to_compare
        self.context_agent_types = context_agent_types
        self.bandit_labels = bandit_labels or [bandit.name for bandit in bandits_to_compare]
        self.decaying_factors = decaying_factors or [1.0]*len(bandits_to_compare)
        self.window_sizes = window_sizes or [5]*len(bandits_to_compare)
        self.bandit_metrics = dict.fromkeys(bandit for bandit in self.bandit_labels)
        self.path_to_save = path_to_save or os.path.join(project_root(), 'mab')
        self.is_visualize = is_visualize
        self.histories = dict.fromkeys([bandit for bandit in self.bandit_labels])
        self.cluster = MiniBatchKMeans(n_clusters=n_clusters)

    def compare_bandits(self, setup_parameters: Callable, initial_population_func: Callable = None) \
            -> Tuple[dict, list]:
        results = dict()
        for i in range(self.launch_num):
            initial_graphs = initial_population_func()
            for j, bandit in enumerate(self.bandits_to_compare):
                optimizer, objective = setup_parameters(initial_graphs=initial_graphs, bandit_type=bandit,
                                                        context_agent_type=self.context_agent_types[j],
                                                        decaying_factor=self.decaying_factors[j],
                                                        window_size=self.window_sizes[j])
                agent = optimizer.mutation.agent
                result = self.launch_bandit(bandit_type=bandit, optimizer=optimizer, objective=objective, bandit_num=j)
                if self.bandit_labels[j] not in results.keys():
                    results[self.bandit_labels[j]] = []
                results[self.bandit_labels[j]].append(result)
        if self.is_visualize:
            self.show_average_action_probabilities(show_action_probabilities=results, actions=agent.actions)
        return results, agent.actions

    def launch_bandit(self, bandit_type: MutationAgentTypeEnum, optimizer: GraphOptimizer, objective: Callable,
                      bandit_num: int):

        stats_action_value_log: Dict[int, List[List[float]]] = dict()

        def log_action_values(next_pop: PopulationT, optimizer: EvoGraphOptimizer):
            values = optimizer.mutation.agent.get_action_values(obs=next_pop[0])
            if 0 not in stats_action_value_log.keys():
                stats_action_value_log[0] = []
            stats_action_value_log[0].append(list(values))
            # MAB agent can be saved here -- commented not to clog up the memory
            # optimizer.mutation.agent.save()

        def log_action_values_with_clusters(next_pop: PopulationT, optimizer: EvoGraphOptimizer):
            obs_contexts = optimizer.mutation.agent.get_context(next_pop)
            self.cluster.partial_fit(obs_contexts)
            centers = self.cluster.cluster_centers_
            for i, center in enumerate(centers):
                values = optimizer.mutation.agent.get_action_values(obs=[center])
                if i not in stats_action_value_log.keys():
                    stats_action_value_log[i] = []
                stats_action_value_log[i].append(list(values))

        # set iteration callback
        if bandit_type == MutationAgentTypeEnum.bandit:
            optimizer.set_iteration_callback(log_action_values)
        elif bandit_type in (MutationAgentTypeEnum.contextual_bandit, MutationAgentTypeEnum.neural_bandit):
            optimizer.set_iteration_callback(log_action_values_with_clusters)
        else:
            raise ValueError("No callback function was specified for that bandit type.")

        found_graphs = optimizer.optimise(objective)
        found_graph = found_graphs[0] if isinstance(found_graphs, Sequence) else found_graphs

        history = optimizer.history
        bandit_label = self.bandit_labels[bandit_num]
        if not self.histories[bandit_label]:
            self.histories[bandit_label] = []
        self.histories[bandit_label].append(history)

        agent = optimizer.mutation.agent
        found_nx_graph = BaseNetworkxAdapter().restore(found_graph)
        final_metrics = objective(found_nx_graph).value

        if not self.bandit_metrics[bandit_label]:
            self.bandit_metrics[bandit_label] = []
        self.bandit_metrics[bandit_label].append(final_metrics)

        print('History of action probabilities:')
        pprint(stats_action_value_log)
        if self.is_visualize:
            self.show_fitness_line(found_nx_graph=found_nx_graph, final_metrics=final_metrics,
                                   history=history)
            self.show_action_probabilities(bandit_type=bandit_type, stats_action_value_log=stats_action_value_log,
                                           actions=agent.actions)

        return stats_action_value_log

    @staticmethod
    def show_fitness_line(found_nx_graph, final_metrics, history):
        draw_graphs_subplots(found_nx_graph, titles=[f'Found Graph (fitness={final_metrics})'])
        history.show.fitness_line()

    def show_action_probabilities(self, bandit_type: MutationAgentTypeEnum, stats_action_value_log,
                                  actions, is_average: bool = False):
        if is_average:
            titles = ['Average action Expectation Values', 'Average action Probabilities']
        else:
            titles = ['Action Expectation Values', 'Action Probabilities']
        if bandit_type == MutationAgentTypeEnum.bandit:
            plot_action_values(stats=stats_action_value_log[0], action_tags=actions, titles=titles)
            plt.show()
        else:
            centers = self.cluster.cluster_centers_
            for i in range(self.cluster.n_clusters):
                if len(centers[i]) > 1:
                    titles_centers = [title + f' for cluster with center idx={i}' for title in titles]
                else:
                    titles_centers = [title + f' for cluster with center {centers[i]}' for title in titles]
                plot_action_values(stats=stats_action_value_log[i], action_tags=actions,
                                   titles=titles_centers)
                plt.show()

    def show_average_action_probabilities(self, show_action_probabilities: dict, actions):
        """ Shows action probabilities across several launches. """
        for idx, bandit in enumerate(list(show_action_probabilities.keys())):
            total_sum = None
            for launch in show_action_probabilities[bandit]:
                if not total_sum:
                    total_sum = launch
                    break
                for cluster in launch.keys():
                    for action_probabilities_list_idx in range(len(total_sum[cluster])):
                        for action_probability_idx in range(len(total_sum[cluster][action_probabilities_list_idx])):
                            if action_probabilities_list_idx >= len(launch[cluster]):
                                continue
                            if action_probability_idx >= len(launch[cluster][action_probabilities_list_idx]):
                                continue
                            total_sum[cluster][action_probabilities_list_idx][action_probability_idx] += \
                                launch[cluster][action_probabilities_list_idx][action_probability_idx]
            for cluster in total_sum.keys():
                for action_probabilities_list_idx in range(len(total_sum[cluster])):
                    for action_probability_idx in range(len(total_sum[cluster][action_probabilities_list_idx])):
                        total_sum[cluster][action_probabilities_list_idx][action_probability_idx] /= \
                            len(show_action_probabilities[bandit])
            self.show_action_probabilities(bandit_type=MutationAgentTypeEnum(self.bandits_to_compare[idx]),
                                           stats_action_value_log=total_sum,
                                           actions=actions,
                                           is_average=True)

    def show_boxplots(self):
        sns.boxplot(data=pd.DataFrame(self.bandit_metrics))
        plt.title('Metrics', fontsize=15)
        plt.show()

    def show_fitness_lines(self):
        multiple_fitness_lines = MultipleFitnessLines(histories_to_compare=self.histories)
        multiple_fitness_lines.visualize()


def setup_parameters(initial_graphs: List[Graph], bandit_type: MutationAgentTypeEnum,
                     context_agent_type: ContextAgentTypeEnum,
                     target_size: int, trial_timeout: float,
                     mutation_types: List[MutationTypesEnum] = None,
                     decaying_factor: float = 1.0,
                     window_size: int = 1):
    objective = Objective({'graph_size': lambda graph: abs(target_size -
                                                           graph.number_of_nodes())})

    # Build the optimizer
    optimizer, _ = graph_search_setup(
        objective=objective,
        optimizer_cls=EvoGraphOptimizer,
        algorithm_parameters=get_graph_gp_params(objective=objective,
                                                 adaptive_mutation_type=bandit_type,
                                                 context_agent_type=context_agent_type,
                                                 decaying_factor=decaying_factor,
                                                 window_size=window_size,
                                                 mutation_types=mutation_types),
        timeout=timedelta(minutes=trial_timeout),
        num_iterations=target_size * 3,
        initial_graphs=initial_graphs
    )
    return optimizer, objective


def initial_population_func(graph_size: List[int] = None, pop_size: int = None, initial_graphs: List[Graph] = None):
    if initial_graphs:
        return initial_graphs
    initial_graphs = [generate_labeled_graph('tree', graph_size[i], directed=True, node_labels=['x'])
                      for i in range(pop_size)]
    initial_opt_graphs = [BaseNetworkxAdapter().adapt(item=graph) for graph in initial_graphs]
    return initial_opt_graphs


if __name__ == '__main__':
    timeout = 0.3
    launch_num = 1
    target_size = 50

    # `bandits_to_compare`, `context_agent_types` and `bandit_labels` correlate one to one.
    # Context must be specified for each bandit: for contextual and neural bandits real context must be specified,
    # for simple bandits -- ContextAgentTypeEnum.none
    bandits_to_compare = [MutationAgentTypeEnum.bandit, MutationAgentTypeEnum.contextual_bandit]
    context_agent_types = [ContextAgentTypeEnum.none_encoding, ContextAgentTypeEnum.operations_quantity]
    bandit_labels = ['simple_bandit', f'context_{context_agent_types[1].name}']

    setup_parameters_func = partial(setup_parameters, target_size=target_size, trial_timeout=timeout)
    initial_population_func = partial(initial_population_func,
                                      graph_size=[random.randint(5, 10) for _ in range(10)] +
                                                 [random.randint(90, 95) for _ in range(10)],
                                      pop_size=20)

    helper = MABSyntheticExperimentHelper(timeout=timeout, launch_num=launch_num, bandits_to_compare=bandits_to_compare,
                                          bandit_labels=bandit_labels, context_agent_types=context_agent_types,
                                          n_clusters=2, is_visualize=True)
    helper.compare_bandits(initial_population_func=initial_population_func,
                           setup_parameters=setup_parameters_func)
    helper.show_boxplots()
    helper.show_fitness_lines()



from abc import ABC
from datetime import timedelta
from typing import Optional, Callable, Dict

import numpy as np
from hyperopt import tpe, hp
from hyperopt.early_stop import no_progress_loss
from hyperopt.pyll import Apply

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.log import default_log
from golem.core.optimisers.objective import ObjectiveFunction
from golem.core.optimisers.timer import Timer
from golem.core.tuning.search_space import SearchSpace, get_node_operation_parameter_label
from golem.core.tuning.tuner_interface import BaseTuner


class HyperoptTuner(BaseTuner, ABC):
    """Base class for hyperparameters optimization based on hyperopt library

    Args:
      objective_evaluate: objective to optimize
      adapter: the function for processing of external object that should be optimized
      search_space: SearchSpace instance
      iterations: max number of iterations
      early_stopping_rounds: Optional max number of stagnating iterations for early stopping. If ``None``, will be set
          to ``max(100, int(np.sqrt(iterations) * 10))``.
      timeout: max time for tuning
      n_jobs: num of ``n_jobs`` for parallelization (``-1`` for use all cpu's)
      deviation: required improvement (in percent) of a metric to return tuned graph.
        By default, ``deviation=0.05``, which means that tuned graph will be returned
        if it's metric will be at least 0.05% better than the initial.
      algo: algorithm for hyperparameters optimization with signature similar to :obj:`hyperopt.tse.suggest`
    """

    def __init__(self, objective_evaluate: ObjectiveFunction,
                 search_space: SearchSpace,
                 adapter: Optional[BaseOptimizationAdapter] = None,
                 iterations: int = 100,
                 early_stopping_rounds: Optional[int] = None,
                 timeout: timedelta = timedelta(minutes=5),
                 n_jobs: int = -1,
                 deviation: float = 0.05,
                 algo: Callable = tpe.suggest):
        early_stopping_rounds = early_stopping_rounds or max(100, int(np.sqrt(iterations) * 10))
        super().__init__(objective_evaluate,
                         search_space,
                         adapter,
                         iterations,
                         early_stopping_rounds,
                         timeout,
                         n_jobs,
                         deviation)

        self.early_stop_fn = no_progress_loss(iteration_stop_count=self.early_stopping_rounds)
        self.max_seconds = int(timeout.seconds) if timeout is not None else None
        self.algo = algo
        self.log = default_log(self)

    def _update_remaining_time(self, tuner_timer: Timer):
        self.max_seconds = self.max_seconds - tuner_timer.minutes_from_start * 60


def get_parameter_hyperopt_space(search_space: SearchSpace,
                                 operation_name: str,
                                 parameter_name: str,
                                 label: str = 'default') -> Optional[Apply]:
    """
    Function return hyperopt object with search_space from search_space dictionary

    Args:
        search_space: SearchSpace with parameters per operation
        operation_name: name of the operation
        parameter_name: name of hyperparameter of particular operation
        label: label to assign in hyperopt pyll

    Returns:
        parameter range
    """

    # Get available parameters for current operation
    operation_parameters = search_space.parameters_per_operation.get(operation_name)

    if operation_parameters is not None:
        parameter_properties = operation_parameters.get(parameter_name)
        hyperopt_distribution = parameter_properties.get('hyperopt-dist')
        sampling_scope = parameter_properties.get('sampling-scope')
        if hyperopt_distribution == hp.loguniform:
            sampling_scope = [np.log(x) for x in sampling_scope]
        return hyperopt_distribution(label, *sampling_scope)
    else:
        return None


def get_node_parameters_for_hyperopt(search_space: SearchSpace, node_id: int, operation_name: str) \
        -> Dict[str, Apply]:
    """
    Function for forming dictionary with hyperparameters of the node operation for the ``HyperoptTuner``

    Args:
        search_space: SearchSpace with parameters per operation
        node_id: number of node in graph.nodes list
        operation_name: name of operation in the node

    Returns:
        parameters_dict: dictionary-like structure with labeled hyperparameters
        and their range per operation
    """

    # Get available parameters for current operation
    parameters_list = search_space.get_parameters_for_operation(operation_name)

    parameters_dict = {}
    for parameter_name in parameters_list:
        node_op_parameter_name = get_node_operation_parameter_label(node_id, operation_name, parameter_name)

        # For operation get range where search can be done
        space = get_parameter_hyperopt_space(search_space, operation_name, parameter_name, node_op_parameter_name)

        parameters_dict.update({node_op_parameter_name: space})

    return parameters_dict



import os
from pathlib import Path
from typing import Optional, Union, Dict, Any, List

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.widgets import Button

from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.paths import default_data_dir
from golem.visualisation.opt_history.fitness_line import with_alternate_matplotlib_backend
from golem.visualisation.opt_history.history_visualization import HistoryVisualization
from golem.visualisation.opt_history.utils import show_or_save_figure


class GraphsInteractive(HistoryVisualization):

    @with_alternate_matplotlib_backend
    def visualize(self, save_path: Optional[Union[os.PathLike, str]] = None, dpi: Optional[int] = None,
                  per_time: Optional[bool] = None, graph_show_kwargs: Optional[Dict[str, Any]] = None):
        """ Shows the structure of the best individuals of all time.
        :param save_path: path to save the visualization. If set, then the image will be saved, and if not,
            it will be displayed.
        :param dpi: DPI of the output figure.
        :param per_time: defines whether to show time grid if it is available in history.
        :param graph_show_kwargs: keyword arguments of `graph.show()` function.
        """

        save_path = save_path or self.get_predefined_value('save_path')
        dpi = dpi or self.get_predefined_value('dpi')
        graph_show_kwargs = graph_show_kwargs or self.visuals_params.get('graph_show_params') or {}

        # fig, axes = plt.subplots(1, 2, figsize=(15, 10))
        # ax_fitness, ax_graph = axes
        fig, ax_graph = plt.subplots(1, 1, figsize=(10, 10))
        ax_graph.axis('off')

        x_template = 'best individual #{}'
        best_individuals = {i: ind
                            for i, ind in enumerate(self.history.archive_history[-1])}

        class InteractivePlot:
            temp_path = Path(default_data_dir(), 'current_graph.png')

            def __init__(self, best_individuals: Dict[int, Individual]):
                self.best_x: List[int] = list(best_individuals.keys())
                self.best_individuals: List[Individual] = list(best_individuals.values())
                self.index: int = len(self.best_individuals) - 1
                # self.time_line = ax_fitness.axvline(self.best_x[self.index], color='r', alpha=0.7)
                self.graph_images: List[np.ndarray] = []
                self.generate_graph_images()
                self.update_graph()

            def generate_graph_images(self):
                for ind in self.best_individuals:
                    graph = ind.graph
                    graph.show(self.temp_path, **graph_show_kwargs)
                    self.graph_images.append(plt.imread(str(self.temp_path)))
                self.temp_path.unlink()

            def update_graph(self):
                ax_graph.imshow(self.graph_images[self.index])
                x = self.best_x[self.index]
                fitness = self.best_individuals[self.index].fitness
                ax_graph.set_title(f'The best graph at {x_template.format(x)}, fitness={fitness}')

            # def update_time_line(self):
            #     self.time_line.set_xdata(self.best_x[self.index])

            def step_index(self, step: int):
                self.index = (self.index + step) % len(self.best_individuals)
                self.update_graph()
                # self.update_time_line()
                plt.draw()

            def next(self, event):
                self.step_index(1)

            def prev(self, event):
                self.step_index(-1)

        callback = InteractivePlot(best_individuals)

        if not save_path:  # display buttons only for an interactive plot
            ax_prev = plt.axes([0.7, 0.05, 0.1, 0.075])
            ax_next = plt.axes([0.81, 0.05, 0.1, 0.075])
            b_next = Button(ax_next, 'Next')
            b_next.on_clicked(callback.next)
            b_prev = Button(ax_prev, 'Previous')
            b_prev.on_clicked(callback.prev)

        show_or_save_figure(fig, save_path, dpi)



from datetime import timedelta
from functools import partial
import random

from typing import List, Any

from examples.adaptive_optimizer.mab_experiment_different_targets import get_graph_gp_params
from examples.synthetic_graph_evolution.generators import generate_labeled_graph
from examples.synthetic_graph_evolution.graph_search import graph_search_setup
from experiments.mab.experiment_with_fickle_mutations.fickle_mutations import fake_add_mutation, fake_add_mutation2, \
    fake_add_mutation3
from experiments.mab.mab_synthetic_experiment_helper import MABSyntheticExperimentHelper
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.dag.graph import Graph
from golem.core.optimisers.adaptive.context_agents import ContextAgentTypeEnum
from golem.core.optimisers.adaptive.operator_agent import MutationAgentTypeEnum
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.objective import Objective


def setup_parameters(initial_graphs: List[Graph], bandit_type: MutationAgentTypeEnum,
                     context_agent_type: ContextAgentTypeEnum,
                     target_size: int, trial_timeout: float,
                     mutation_types: List[Any] = None,
                     decaying_factor: float = 1.0,
                     window_size: int = 5):
    objective = Objective({'graph_size': lambda graph: abs(target_size -
                                                           graph.number_of_nodes())})

    # Build the optimizer
    optimizer, _ = graph_search_setup(
        objective=objective,
        optimizer_cls=EvoGraphOptimizer,
        algorithm_parameters=get_graph_gp_params(objective=objective,
                                                 adaptive_mutation_type=bandit_type,
                                                 context_agent_type=context_agent_type,
                                                 decaying_factor=decaying_factor,
                                                 window_size=window_size,
                                                 mutation_types=mutation_types),
        timeout=timedelta(minutes=trial_timeout),
        num_iterations=target_size * 3,
        initial_graphs=initial_graphs,
        node_types=['x', 'y', 'z', 'w', 'v']
    )
    return optimizer, objective


def initial_population_func(graph_size: List[int] = None, pop_size: int = None, initial_graphs: List[Graph] = None):
    if initial_graphs:
        return initial_graphs
    initial_graphs = [generate_labeled_graph('tree', graph_size[i], directed=True, node_labels=['x'])
                      for i in range(pop_size)]
    initial_opt_graphs = [BaseNetworkxAdapter().adapt(item=graph) for graph in initial_graphs]
    return initial_opt_graphs


if __name__ == '__main__':
    timeout = 3.5
    launch_num = 3
    target_size = 200

    # `bandits_to_compare`, `context_agent_types` and `bandit_labels` correlate one to one.
    # Context must be specified for each bandit: for contextual and neural bandits real context must be specified,
    # for simple bandits -- ContextAgentTypeEnum.none
    bandits_to_compare = [MutationAgentTypeEnum.contextual_bandit, MutationAgentTypeEnum.contextual_bandit]
    context_agent_types = [ContextAgentTypeEnum.nodes_num, ContextAgentTypeEnum.nodes_num]
    bandit_labels = ['bandit_base', 'bandit_3_0.6']
    decaying_factors = [1.0, 0.6]
    window_sizes = [1, 3]
    mutation_types = [
        fake_add_mutation,
        fake_add_mutation2,
        fake_add_mutation3
    ]

    setup_parameters_func = partial(setup_parameters, target_size=target_size, trial_timeout=timeout)
    initial_population_func = partial(initial_population_func,
                                      graph_size=[random.randint(3, 5) for _ in range(20)],
                                      pop_size=20)

    helper = MABSyntheticExperimentHelper(timeout=timeout, launch_num=launch_num, bandits_to_compare=bandits_to_compare,
                                          bandit_labels=bandit_labels, context_agent_types=context_agent_types,
                                          n_clusters=1, is_visualize=False, decaying_factors=decaying_factors,
                                          window_sizes=window_sizes)
    results, actions = helper.compare_bandits(initial_population_func=initial_population_func,
                                              setup_parameters=setup_parameters_func)
    helper.show_boxplots()
    helper.show_fitness_lines()
    helper.show_average_action_probabilities(results, actions)



from copy import deepcopy
from inspect import signature
from typing import Any, Dict, Type, TypeVar, Callable


INSTANCE_OR_CALLABLE = TypeVar('INSTANCE_OR_CALLABLE', object, Callable)


def any_to_json(obj: INSTANCE_OR_CALLABLE) -> Dict[str, Any]:
    return {k: v for k, v in sorted(vars(obj).items()) if not _is_log_var(k)}


def any_from_json(cls: Type[INSTANCE_OR_CALLABLE], json_obj: Dict[str, Any]) -> INSTANCE_OR_CALLABLE:
    obj = cls.__new__(cls)
    vars(obj).update(json_obj)
    return obj


def _is_log_var(varname: str) -> bool:
    return varname.strip('_').startswith('log')



from examples.adaptive_optimizer.experiment_setup import run_adaptive_mutations_with_context
from examples.adaptive_optimizer.mab_experiment_different_targets import run_experiment_node_num, \
    run_experiment_edge_num, run_experiment_graphs_ratio_edges_nodes, run_experiment_trees
from golem.core.optimisers.adaptive.context_agents import ContextAgentTypeEnum
from golem.core.optimisers.adaptive.operator_agent import MutationAgentTypeEnum


if __name__ == '__main__':
    """Run adaptive optimizer on different targets to see how neural multi-armed bandit agent converges
    to different probabilities of actions (i.e. mutations) for different targets."""
    adaptive_mutation_type = MutationAgentTypeEnum.contextual_bandit
    context_agent_type = ContextAgentTypeEnum.nodes_num

    run_experiment_node_num(trial_timeout=2,
                            adaptive_mutation_type=adaptive_mutation_type,
                            context_agent_type=context_agent_type,
                            run_func=run_adaptive_mutations_with_context)
    run_experiment_edge_num(trial_timeout=2,
                            adaptive_mutation_type=adaptive_mutation_type,
                            context_agent_type=context_agent_type,
                            run_func=run_adaptive_mutations_with_context)
    run_experiment_trees(trial_timeout=10, trial_iterations=2000,
                         adaptive_mutation_type=adaptive_mutation_type,
                         context_agent_type=context_agent_type,
                         run_func=run_adaptive_mutations_with_context)
    run_experiment_graphs_ratio_edges_nodes(trial_timeout=10, trial_iterations=2000,
                                            adaptive_mutation_type=adaptive_mutation_type,
                                            context_agent_type=context_agent_type,
                                            run_func=run_adaptive_mutations_with_context)



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Dict, Any

from golem.core.log import default_log
from golem.visualisation.opt_history.arg_constraint_wrapper import ArgConstraintWrapper

if TYPE_CHECKING:
    from golem.core.optimisers.opt_history_objects.opt_history import OptHistory


class HistoryVisualization(metaclass=ArgConstraintWrapper):
    """ Base class for creating visualizations of the optimization history.
    The only necessary method is 'visualize' - it must show or save the plot in any form after the call.

    One should refer the OptHistory instance as `self.history` to be able to connect one's visualization
    to `OptHistory.show()`. See the examples of connecting visualizations in the module `opt_viz.py`.

    It is good practice to be aware of constraints on your visualizations. You can either implement
    default constraints that will catch your kwarg across all the visualizations or define your single
    class specific constraints by assigning them to `constraint_checkers` class attribute.
    See `golem.core.visualisation.opt_history.arg_constraint_wrapper.py` for examples.
    """
    constraint_checkers = []  # Use this for class-specific constraint checkers.

    def __init__(self, history: 'OptHistory', visuals_params: Dict[str, Any] = None):
        self.visuals_params = visuals_params or {}
        self.history = history
        self.log = default_log(self)

    @abstractmethod
    def visualize(self, *args, **kwargs):
        raise NotImplementedError()

    def get_predefined_value(self, param: str):
        return self.visuals_params.get(param)



SULFUR_DEFAULT_VALENCE = 6
MIN_LONG_CYCLE_SIZE = 6

# normalization constants, statistics from 250k_rndm_zinc_drugs_clean.smi dataset
ZINC_LOGP_MEAN = 2.4570953396190123
ZINC_LOGP_STD = 1.434324401111988
ZINC_SA_MEAN = -3.0525811293166134
ZINC_SA_STD = 0.8335207024513095
ZINC_CYCLE_MEAN = -0.0485696876403053
ZINC_CYCLE_STD = 0.2860212110245455



import cProfile
import os
import pstats
import subprocess

from golem.utilities.requirements_notificator import warn_requirement

try:
    import gprof2dot
except ImportError:
    warn_requirement('gprof2dot', 'golem[profilers]', should_raise=True)
try:
    import snakeviz
except ImportError:
    warn_requirement('snakeviz', 'golem[profilers]', should_raise=True)


class TimeProfiler:
    """Profile code and visual interpret results of it
    """

    def __init__(self):
        self.folder = None
        self.profiler = cProfile.Profile()
        self.profiler.enable()

    def _generate_pstats(self, path: str):
        """Aggregate profiler statistics and create :obj:`output.pstats` from ``Profiler``

        Args:
            path: path to save results
        """

        self.path_stats = os.path.join(os.path.abspath(path), 'output.pstats')
        stats = pstats.Stats(self.profiler)
        stats.dump_stats(self.path_stats)

    def profile(self, path: str, node_percent: float = 0.5, edge_percent: float = 0.1, open_web: bool = False):
        """Method to convert the statistics from profiler to visual representation

        Args:
            spath: path to save profiling result
            node_percent: eliminate nodes below this threshold [default: 0.5]
            edge_percent: eliminate edges below this threshold [default: 0.1]
            open_web: boolean parametr to open web-interface
        """

        self.profiler.disable()

        self.folder = os.path.abspath(path)

        if not os.path.exists(self.folder):
            os.mkdir(self.folder)

        self._generate_pstats(self.folder)

        # Creating the PNG call-graph with the cumulative time.
        path = os.path.join(self.folder, 'graph.png')
        subprocess.getstatusoutput(f"gprof2dot -n{node_percent} -e{edge_percent} -s -f pstats {self.path_stats} | "
                                   f"dot -Tpng -o {path}")

        # Opening the web interface to view the detailed profiler stats.
        if open_web:
            subprocess.getstatusoutput(f"snakeviz {self.path_stats}")



import os
from abc import abstractmethod
from datetime import timedelta, datetime
from io import StringIO
from itertools import product
from typing import Optional, Type, Sequence, List, Callable, Union, Tuple

import numpy as np
import pandas as pd
from networkx import DiGraph

from examples.synthetic_graph_evolution.generators import generate_labeled_graph
from examples.synthetic_graph_evolution.utils import draw_graphs_subplots
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.dag.graph import Graph
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.opt_history_objects.opt_history import OptHistory
from golem.core.optimisers.optimizer import GraphOptimizer
from golem.metrics.graph_metrics import nxgraph_stats


class BaseExperimentLauncher:
    """ Base class for all experiment launchers. Setups the logic of saving experiments results as
    it is required in ExperimentAnalyzer. """
    def __init__(self, path_to_save: str, is_save_visualizations: bool = True,
                 optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer, num_trials: int = 10,
                 trial_timeout: Optional[int] = None, trial_iterations: Optional[int] = None, **kwargs):
        self.optimizer_cls = optimizer_cls
        self.num_trials = num_trials
        self.trial_timeout = trial_timeout
        self.trial_iterations = trial_iterations
        self.path_to_save = path_to_save
        self.is_save_visualizations = is_save_visualizations

    @abstractmethod
    def launch(self, optimizer_setup: Callable, **kwargs):
        """
         Launches experiments for specified number of trials.
         :param optimizer_setup: function that setups all infrastructure for launches.
         """
        raise NotImplementedError()

    @abstractmethod
    def _launch_experiment(self, optimizer_setup: Callable, **kwargs):
        """ Launches one experiment with specified params. """
        raise NotImplementedError()

    @staticmethod
    def _save_experiment_results(path_to_save: str, optimizer: GraphOptimizer):
        """ Saves experiment result as it is required in ExperimentAnalyzer. """
        os.makedirs(path_to_save, exist_ok=True)

        # save metrics
        obj_names = optimizer.objective.metric_names
        fitness = dict.fromkeys(obj_names)
        for ind in optimizer.best_individuals:
            for j, metric in enumerate(obj_names):
                if not fitness[metric]:
                    fitness[metric] = []
                fitness[metric].append(ind.fitness.values[j])
        df_metrics = pd.DataFrame(fitness)
        df_metrics.to_csv(os.path.join(path_to_save, 'metrics.csv'))

        # save history
        history = optimizer.history
        history.save(os.path.join(path_to_save, 'history.json'))

    @staticmethod
    def _save_visualizations(history: OptHistory, setup_name: str, path_to_save: str):
        """ Saves visualizations of results. """
        path_to_save = os.path.join(path_to_save, 'visualizations')
        os.makedirs(path_to_save, exist_ok=True)
        diversity_path_to_save = os.path.join(path_to_save, 'diversity')
        os.makedirs(diversity_path_to_save, exist_ok=True)
        diversity_filename = f'diversity_hist_{setup_name}.gif'
        history.show.diversity_population(save_path=os.path.join(diversity_path_to_save, diversity_filename))
        history.show.diversity_line(save_path=diversity_path_to_save, show=False)
        history.show.fitness_line(save_path=os.path.join(path_to_save, 'fitness_line.png'))


class GraphStructureExperimentLauncher(BaseExperimentLauncher):
    """
    Class that allows to easily set up experiments and save results in format
    required for ExperimentAnalyzer.
    GraphStructureExperimentLauncher provide common logic for experiments concerned with growing
    a graph with specified structure and type.
    The format is:
        setup (e.g. configuration of framework)
        \
         dataset
               \
               launch
                     \
                      all collected data (metrics.csv, history, visualizations, etc)
    Example of usage: examples/experiment_launcher/experiment_launcher.py
    """
    def __init__(self, graph_names: List[str], graph_sizes: List[int], path_to_save: str,
                 is_save_visualizations: bool = True, optimizer_cls: Type[GraphOptimizer] = EvoGraphOptimizer,
                 node_types: Optional[Sequence[str]] = None, num_trials: int = 10, trial_timeout: Optional[int] = None,
                 trial_iterations: Optional[int] = None):
        super().__init__(path_to_save=path_to_save, is_save_visualizations=is_save_visualizations,
                         optimizer_cls=optimizer_cls, num_trials=num_trials, trial_timeout=trial_timeout,
                         trial_iterations=trial_iterations)
        self.graph_names = graph_names
        self.graph_sizes = graph_sizes
        self.node_types = node_types

    def launch(self, optimizer_setup: Callable, **kwargs):
        """
        Launches experiments for product of all graph names and graph sizes for specified number of trials.
        :param optimizer_setup: function that setups all infrastructure for launches.
        """
        log = StringIO()
        if not self.node_types:
            self.node_types = ['X']
        for graph_name, num_nodes in product(self.graph_names, self.graph_sizes):
            experiment_id = f'Experiment [graph={graph_name} graph_size={num_nodes}]'
            trial_results = []
            for i in range(self.num_trials):
                setup_name = optimizer_setup.__name__
                cur_path_to_save = os.path.join(self.path_to_save, setup_name, f'{graph_name}_{num_nodes}', str(i))
                os.makedirs(cur_path_to_save, exist_ok=True)
                start_time = datetime.now()
                print(f'\nTrial #{i} of {experiment_id} started at {start_time}', file=log)

                optimizer, objective, target_graph = self._launch_experiment(graph_name=graph_name,
                                                                             num_nodes=num_nodes,
                                                                             node_types=self.node_types,
                                                                             optimizer_setup=optimizer_setup)

                found_graphs = optimizer.optimise(objective)
                found_graph = found_graphs[0] if isinstance(found_graphs, Sequence) else found_graphs
                history = optimizer.history

                trial_results.extend(history.final_choices)
                found_nx_graph = BaseNetworkxAdapter().restore(found_graph)

                duration = datetime.now() - start_time
                print(f'Trial #{i} finished, spent time: {duration}', file=log)
                print('target graph stats: ', nxgraph_stats(target_graph), file=log)
                print('found graph stats: ', nxgraph_stats(found_nx_graph), file=log)

                self._save_experiment_results(path_to_save=cur_path_to_save, optimizer=optimizer)

                if self.is_save_visualizations:
                    self._save_visualizations(target_graph=target_graph, found_nx_graph=found_nx_graph,
                                              history=history, setup_name=setup_name, path_to_save=cur_path_to_save)

            # Compute mean & std for metrics of trials
            ff = objective.format_fitness
            trial_metrics = np.array([ind.fitness.values for ind in trial_results])
            trial_metrics_mean = trial_metrics.mean(axis=0)
            trial_metrics_std = trial_metrics.std(axis=0)
            print(f'{experiment_id} finished with metrics:\n'
                  f'mean={ff(trial_metrics_mean)}\n'
                  f'std={ff(trial_metrics_std)}',
                  file=log)
            print(log.getvalue())

    def _launch_experiment(self, optimizer_setup: Callable, **kwargs) \
            -> Tuple[GraphOptimizer, Objective, Union[Graph, DiGraph]]:
        """ Experiment launch for structure search task. """
        graph_name = kwargs['graph_name']
        num_nodes = kwargs['num_nodes']
        node_types = kwargs['node_types']
        # Generate random target graph and run the optimizer
        target_graph = generate_labeled_graph(graph_name, num_nodes, node_types)
        target_graph = target_graph.reverse()
        # Run optimizer setup
        optimizer, objective = optimizer_setup(target_graph=target_graph,
                                               optimizer_cls=self.optimizer_cls,
                                               node_types=node_types,
                                               timeout=timedelta(minutes=self.trial_timeout)
                                               if self.trial_timeout else None,
                                               num_iterations=self.trial_iterations)
        return optimizer, objective, target_graph

    @staticmethod
    def _save_visualizations(history: OptHistory, setup_name: str, path_to_save: str, **kwargs):
        """ Saves visualizations of results. """
        target_graph = kwargs['target_graph']
        found_nx_graph = kwargs['found_nx_graph']
        BaseExperimentLauncher._save_visualizations(history=history, setup_name=setup_name, path_to_save=path_to_save)
        draw_graphs_subplots(target_graph, found_nx_graph,
                             titles=['Target Graph', 'Found Graph'], show=False, path_to_save=path_to_save)



import os
from typing import Tuple, Set, Optional

import networkx as nx
from rdkit.Chem import GetPeriodicTable
from rdkit.Chem.rdchem import Atom, RWMol
from sphinx.util import requests

from examples.molecule_search.constants import SULFUR_DEFAULT_VALENCE
from examples.molecule_search.mol_graph import MolGraph
from golem.core.log import default_log


def get_default_valence(atom_type: str) -> int:
    if atom_type == "S":
        default_valence = SULFUR_DEFAULT_VALENCE
    else:
        default_valence = GetPeriodicTable().GetDefaultValence(atom_type)
    return default_valence


def get_free_electrons_num(atom: Atom) -> int:
    atom.UpdatePropertyCache()
    atom_type = atom.GetSymbol()
    default_valence = get_default_valence(atom_type)
    explicit_valence = atom.GetExplicitValence()
    free_electrons = default_valence - explicit_valence
    return free_electrons


def get_functional_group(mol_graph: MolGraph, bridge: Tuple[int, int]) -> Set[int]:
    disconnected_graph = mol_graph.get_nx_graph()
    disconnected_graph.remove_edge(*bridge)
    group = nx.node_connected_component(disconnected_graph, bridge[1])
    return group


def largest_ring_size(rw_molecule: RWMol) -> int:
    largest_cycle_len = 0
    cycle_list = rw_molecule.GetRingInfo().AtomRings()
    if cycle_list:
        largest_cycle_len = max(map(len, cycle_list))
    return largest_cycle_len


def download_from_github(save_path: str, github_url: str, message: Optional[str] = None):
    """ Checks if the file exists. If not downloads the file from specified url."""
    save_dir = os.path.dirname(save_path)
    os.makedirs(save_dir, exist_ok=True)

    message = message or f"Downloading a file from {github_url} to {save_dir}..."

    if not os.path.exists(save_path):
        default_log().message(message)
        response = requests.get(github_url)
        with open(save_path, "wb") as new_file:
            new_file.write(response.content)



from itertools import chain
from typing import Any, Dict, List, Sequence, Type, Union

from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.opt_history_objects.individual import Individual
from golem.core.optimisers.opt_history_objects.generation import Generation
from golem.core.optimisers.opt_history_objects.opt_history import OptHistory
from .. import any_from_json, any_to_json
from ...core.optimisers.objective.objective import ObjectiveInfo

MISSING_INDIVIDUAL_ARGS = {
    'metadata': {'MISSING_INDIVIDUAL': 'This individual could not be restored during `OptHistory.load()`'}
}


def _flatten_generations_list(generations_list: List[List[Individual]]) -> List[Individual]:
    def extract_intermediate_parents(ind: Individual):
        for parent in ind.parents:
            if not parent.has_native_generation:
                parents_map[parent.uid] = parent
                extract_intermediate_parents(parent)

    generations_map = {ind.uid: ind for ind in chain(*generations_list)}
    parents_map = {}
    for individual in generations_map.values():
        extract_intermediate_parents(individual)
    parents_map.update(generations_map)
    individuals_pool = list(parents_map.values())
    return individuals_pool


def _generations_list_to_uids(generations_list: List[Generation]) -> List[Generation]:
    new_list = []
    for generation in generations_list:
        new_gen = generation.copy()
        new_gen.data = [individual.uid for individual in generation]
        new_list.append(new_gen)
    return new_list


def _archive_to_uids(generations_list: List[List[Individual]]) -> List[List[str]]:
    return [[individual.uid for individual in generation] for generation in generations_list]


def opt_history_to_json(obj: OptHistory) -> Dict[str, Any]:
    serialized = any_to_json(obj)
    serialized['individuals_pool'] = _flatten_generations_list(serialized['_generations'])
    serialized['_generations'] = _generations_list_to_uids(serialized['_generations'])
    serialized['archive_history'] = _archive_to_uids(serialized['archive_history'])
    return serialized


def _uids_to_individuals(uid_sequence: Sequence[Union[str, Individual]],
                         uid_to_individual_map: Dict[str, Individual]) -> List[Individual]:
    def get_missing_individual(uid: str) -> Individual:
        individual = Individual(OptGraph(), **MISSING_INDIVIDUAL_ARGS, uid=uid)
        return individual

    def uid_to_individual_mapper(uid: Union[str, Individual]) -> Individual:
        return uid_to_individual_map.get(uid, get_missing_individual(uid)) if isinstance(uid, str) else uid

    return list(map(uid_to_individual_mapper, uid_sequence))


def _deserialize_generations_list(generations_list: List[Union[Generation, List[Union[str, Individual]]]],
                                  uid_to_individual_map: Dict[str, Individual]):
    """The operation is executed in-place"""
    for gen_num, generation in enumerate(generations_list):
        individuals = _uids_to_individuals(uid_sequence=generation,
                                           uid_to_individual_map=uid_to_individual_map)
        if isinstance(generations_list[gen_num], Generation):
            generations_list[gen_num].data = individuals
        else:
            generations_list[gen_num] = individuals


def _deserialize_parent_individuals(individuals: List[Individual],
                                    uid_to_individual_map: Dict[str, Individual]):
    """The operation is executed in-place"""

    def deserialize_intermediate_parents(ind):
        parent_op = ind.parent_operator
        if not parent_op:
            return
        parent_individuals = _uids_to_individuals(uid_sequence=parent_op.parent_individuals,
                                                  uid_to_individual_map=uid_to_individual_map)
        object.__setattr__(parent_op, 'parent_individuals', tuple(parent_individuals))
        for parent in parent_individuals:
            if any(isinstance(i, str) for i in parent.parents):
                deserialize_intermediate_parents(parent)

    for individual in individuals:
        deserialize_intermediate_parents(individual)


def opt_history_from_json(cls: Type[OptHistory], json_obj: Dict[str, Any]) -> OptHistory:
    # backward compatibility with history._is_multi_objective field
    if '_is_multi_objective' in json_obj:
        json_obj['_objective'] = ObjectiveInfo(json_obj['_is_multi_objective'])
        del json_obj['_is_multi_objective']

    # OptHistory.individuals are now OptHistory._generations
    if 'individuals' in json_obj:
        json_obj['_generations'] = json_obj.pop('individuals')

    history = any_from_json(cls, json_obj)
    # Read all individuals from history.
    individuals_pool = history.individuals_pool
    uid_to_individual_map = {ind.uid: ind for ind in individuals_pool}
    # The attributes `individuals` and `archive_history` at the moment contain uid strings that must be converted
    # to `Individual` instances.
    _deserialize_generations_list(history.generations, uid_to_individual_map)
    _deserialize_generations_list(history.archive_history, uid_to_individual_map)
    # Process histories with zero generations.
    if len(history.generations) > 0:
        # Process older histories to wrap generations into the new class.
        if isinstance(history.generations[0], list):
            history.generations = [Generation(gen, gen_num) for gen_num, gen in enumerate(history.generations)]
    # Deserialize parents for all generations.
    _deserialize_parent_individuals(list(chain(*history.generations)), uid_to_individual_map)
    # The attribute is used only for serialization.
    del history.individuals_pool
    return history



import json
import os.path
from datetime import datetime
from typing import List, Optional, Union

from golem.core.dag.graph import Graph
from golem.core.log import default_log
from golem.core.paths import project_root
from golem.utilities.serializable import Serializable
from golem.serializers import Serializer
from golem.structural_analysis.graph_sa.results.deletion_sa_approach_result import DeletionSAApproachResult
from golem.structural_analysis.graph_sa.results.object_sa_result import ObjectSAResult, \
    StructuralAnalysisResultsRepository
from golem.structural_analysis.graph_sa.results.utils import EntityTypesEnum


class SAAnalysisResults(Serializable):
    """ Class presenting results of Structural Analysis for the whole graph. """

    def __init__(self):
        self.results_per_iteration = {}
        self._add_empty_iteration_results()
        self.log = default_log('sa_results')

    def _add_empty_iteration_results(self):
        last_iter_num = int(list(self.results_per_iteration.keys())[-1]) if self.results_per_iteration.keys() else -1
        self.results_per_iteration.update({(last_iter_num + 1): self._init_iteration_result()})

    @staticmethod
    def _init_iteration_result() -> dict:
        return {EntityTypesEnum.node.value: [], EntityTypesEnum.edge.value: []}

    @property
    def is_empty(self) -> bool:
        """ Bool value indicating is there any calculated results. """
        if self.results_per_iteration[0] is None and \
                self.results_per_iteration[0] is None:
            return True
        return False

    def get_info_about_worst_result(self, metric_idx_to_optimize_by: int, iter: Optional[int] = None) -> dict:
        """ Returns info about the worst result.
        :param metric_idx_to_optimize_by: metric idx to optimize by
        :param iter: iteration on which to search for. """
        worst_value = None
        worst_result = None
        if iter is None:
            iter = list(self.results_per_iteration.keys())[-1]

        nodes_results = self.results_per_iteration[iter][EntityTypesEnum.node.value]
        edges_results = self.results_per_iteration[iter][EntityTypesEnum.edge.value]

        for i, res in enumerate(nodes_results + edges_results):
            cur_res = res.get_worst_result_with_names(
                metric_idx_to_optimize_by=metric_idx_to_optimize_by)
            if not worst_value or cur_res['value'] > worst_value:
                worst_value = cur_res['value']
                worst_result = cur_res
        return worst_result

    def add_results(self, results: List[ObjectSAResult]):
        if not results:
            return
        key = results[0].entity_type.value
        iter_num = self._get_last_empty_iter(key=key)
        for result in results:
            self.results_per_iteration[iter_num][key].append(result)

    def _get_last_empty_iter(self, key: str) -> int:
        """ Returns number of last iteration with empty key field. """
        for i, result in enumerate(self.results_per_iteration.values()):
            if not result[key]:
                return i
        self._add_empty_iteration_results()
        return list(self.results_per_iteration.keys())[-1]

    def save(self, path: str = None, datetime_in_path: bool = True) -> dict:
        """ Saves SA results in json format. """
        dict_results = dict()
        for iter in self.results_per_iteration.keys():
            dict_results[iter] = {}
            iter_result = self.results_per_iteration[iter]
            for entity_type in iter_result.keys():
                if entity_type not in dict_results[iter].keys():
                    dict_results[iter][entity_type] = {}
                for entity in iter_result[entity_type]:
                    dict_results[iter][entity_type].update(entity.get_dict_results())

        json_data = json.dumps(dict_results, cls=Serializer)

        if not path:
            path = os.path.join(project_root(), 'sa', 'sa_results.json')
        if datetime_in_path:
            file_name = os.path.basename(path).split('.')[0]
            file_name = f"{datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}_{file_name}.json"
            path = os.path.join(os.path.dirname(path), 'sa')
            if not os.path.exists(path):
                os.makedirs(path)
            path = os.path.join(path, file_name)

        with open(path, 'w', encoding='utf-8') as f:
            f.write(json_data)
            self.log.debug(f'SA results saved in the path: {path}.')

        return dict_results

    @staticmethod
    def load(source: Union[str, dict], graph: Optional[Graph] = None) -> 'SAAnalysisResults':
        """ Loads SA results from json format. """
        if isinstance(source, str):
            source = json.load(open(source))

        sa_result = SAAnalysisResults()
        results_repo = StructuralAnalysisResultsRepository()

        for iter in source:
            for entity_type in source[iter]:
                type_list = []
                for entity_idx in source[iter][entity_type]:
                    cur_result = ObjectSAResult(entity_idx=entity_idx,
                                                entity_type=EntityTypesEnum(entity_type))
                    dict_results = source[iter][entity_type][entity_idx]
                    for approach in dict_results:
                        app = results_repo.get_class_by_str(approach)()
                        if isinstance(app, DeletionSAApproachResult):
                            app.add_results(metrics_values=dict_results[approach])
                        else:
                            for entity_to_replace_to in dict_results[approach]:
                                app.add_results(entity_to_replace_to=entity_to_replace_to,
                                                metrics_values=dict_results[approach][entity_to_replace_to])
                        cur_result.add_result(app)
                    type_list.append(cur_result)
                sa_result.add_results(type_list)

        return sa_result





class BaseSAApproachResult:
    """ Base class for presenting all result classes.
    Specifies the main logic of setting and getting calculated metrics. """

    def get_worst_result(self, metric_idx_to_optimize_by: int) -> float:
        """ Returns the worst result among all metrics. """
        raise NotImplementedError()

    def get_worst_result_with_names(self, metric_idx_to_optimize_by: int) -> dict:
        """ Returns the worst result with additional info. """
        raise NotImplementedError()

    def add_results(self, **kwargs):
        """ Adds newly calculated results. """
        raise NotImplementedError

    def get_dict_results(self) -> dict:
        """ Returns dict representation of results. """
        raise NotImplementedError()



from golem.structural_analysis.graph_sa.edge_sa_approaches import EdgeDeletionAnalyze, EdgeReplaceOperationAnalyze
from golem.structural_analysis.graph_sa.node_sa_approaches import NodeDeletionAnalyze, NodeReplaceOperationAnalyze, \
    SubtreeDeletionAnalyze
from golem.structural_analysis.graph_sa.postproc_methods import nodes_deletion, nodes_replacement, subtree_deletion, \
    edges_deletion, edges_replacement

NODE_DELETION = 'NodeDeletionAnalyze'
NODE_REPLACEMENT = 'NodeReplaceOperationAnalyze'
SUBTREE_DELETION = 'SubtreeDeletionAnalyze'
EDGE_DELETION = 'EdgeDeletionAnalyze'
EDGE_REPLACEMENT = 'EdgeReplaceOperationAnalyze'


class StructuralAnalysisApproachesRepository:
    approaches_dict = {NODE_DELETION: {'approach': NodeDeletionAnalyze,
                                       'postproc_method': nodes_deletion},
                       NODE_REPLACEMENT: {'approach': NodeReplaceOperationAnalyze,
                                          'postproc_method': nodes_replacement},
                       SUBTREE_DELETION: {'approach': SubtreeDeletionAnalyze,
                                          'postproc_method': subtree_deletion},
                       EDGE_DELETION: {'approach': EdgeDeletionAnalyze,
                                       'postproc_method': edges_deletion},
                       EDGE_REPLACEMENT: {'approach': EdgeReplaceOperationAnalyze,
                                          'postproc_method': edges_replacement}}

    def approach_by_name(self, approach_name: str):
        return self.approaches_dict[approach_name]['approach']

    def postproc_method_by_name(self, approach_name: str):
        return self.approaches_dict[approach_name]['postproc_method']



from datetime import timedelta
from functools import partial

from examples.synthetic_graph_evolution.generators import generate_labeled_graph
from examples.synthetic_graph_evolution.utils import draw_graphs_subplots
from golem.core.adapter.nx_adapter import BaseNetworkxAdapter
from golem.core.dag.verification_rules import DEFAULT_DAG_RULES
from golem.core.optimisers.genetic.gp_optimizer import EvoGraphOptimizer
from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.base_mutations import MutationTypesEnum
from golem.core.optimisers.genetic.operators.crossover import CrossoverTypesEnum
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.objective import Objective
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.core.optimisers.optimizer import GraphGenerationParams
from golem.metrics.edit_distance import tree_edit_dist


def run_graph_search(size=16, timeout=8, visualize=True):
    # Generate target graph that will be sought by optimizer
    node_types = ('a', 'b')
    target_graph = generate_labeled_graph('tree', size, node_labels=node_types)

    # Generate initial population with small tree graphs
    initial_graphs = [generate_labeled_graph('tree', 5, node_types) for _ in range(10)]
    # Setup objective: edit distance to target graph
    objective = Objective(partial(tree_edit_dist, target_graph))

    # Setup optimization parameters
    requirements = GraphRequirements(
        early_stopping_iterations=100,
        timeout=timedelta(minutes=timeout),
        n_jobs=-1,
    )
    gp_params = GPAlgorithmParameters(
        genetic_scheme_type=GeneticSchemeTypesEnum.parameter_free,
        max_pop_size=50,
        mutation_types=[MutationTypesEnum.single_add,
                        MutationTypesEnum.single_drop,
                        MutationTypesEnum.single_change],
        crossover_types=[CrossoverTypesEnum.subtree]
    )
    graph_gen_params = GraphGenerationParams(
        adapter=BaseNetworkxAdapter(),  # Example works with NetworkX graphs
        rules_for_constraint=DEFAULT_DAG_RULES,  # We don't want cycles in the graph
        available_node_types=node_types  # Node types that can appear in graphs
    )
    all_parameters = (requirements, graph_gen_params, gp_params)

    # Build and run the optimizer
    optimiser = EvoGraphOptimizer(objective, initial_graphs, *all_parameters)
    found_graphs = optimiser.optimise(objective)
    if visualize:
        # Restore the NetworkX graph back from internal Graph representation
        found_graph = graph_gen_params.adapter.restore(found_graphs[0])
        draw_graphs_subplots(target_graph, found_graph, titles=['Target Graph', 'Found Graph'])
        optimiser.history.show.fitness_line()
    return found_graphs


if __name__ == '__main__':
    """
    In this example Optimizer is expected to find the target graph
    using Tree Edit Distance metric and a random tree (nx.random_tree) as target.
    The convergence can be seen from achieved metrics and visually from graph plots.
    """
    run_graph_search(visualize=True)



# This code is modified part of DEAP library (Library URL: https://github.com/DEAP/deap).
from bisect import bisect_right
from operator import eq
from typing import Callable, Optional

from golem.core.optimisers.genetic.operators.operator import PopulationT
from golem.core.optimisers.opt_history_objects.individual import Individual


class HallOfFame:
    """
    The hall of fame contains the best individual that ever lived in the
    population during the evolution. It is lexicographically sorted at all
    time so that the first element of the hall of fame is the individual that
    has the best first fitness value ever seen, according to the weights
    provided to the fitness at creation time.

    The insertion is made so that old individuals have priority on new
    individuals. A single copy of each individual is kept at all time, the
    equivalence between two individuals is made by the operator passed to the
    *similar* argument.

    :param maxsize: The maximum number of individual to keep in the hall of
                    fame.
    :param similar: An equivalence operator between two individuals, optional.
                    It defaults to operator :func:`operator.eq`.

    The class :class:`HallOfFame` provides an interface similar to a list
    (without being one completely). It is possible to retrieve its length, to
    iterate on it forward and backward and to get an item or a slice from it.
    """

    def __init__(self, maxsize: Optional[int], similar: Callable = eq):
        self.maxsize = maxsize or 0
        self.keys = list()
        self.items = list()
        self.similar = similar

    def update(self, population: PopulationT):
        """
        Update the hall of fame with the *population* by replacing the
        worst individuals in it by the best individuals present in
        *population* (if they are better). The size of the hall of fame is
        kept constant.

        :param population: A list of individual with a fitness attribute to
                           update the hall of fame with.
        """
        if not population:
            return
        for ind in population:
            if len(self) == 0 and self.maxsize != 0:
                # Working on an empty hall of fame is problematic for the loop
                self.insert(population[0])
                continue
            if ind.fitness > self[-1].fitness or len(self) < self.maxsize:
                for hofer in self:
                    # Loop through the hall of fame to check for any similar individual
                    if self.similar(ind, hofer):
                        break
                else:
                    # The individual is unique and strictly better than the worst
                    if len(self) >= self.maxsize:
                        self.remove(-1)
                    self.insert(ind)

    def insert(self, item: Individual):
        """
        Insert a new individual in the hall of fame using the
        :func:`~bisect.bisect_right` function. The inserted individual is
        inserted on the right side of an equal individual. Inserting a new
        individual in the hall of fame also preserve the hall of fame's order.
        This method **does not** check for the size of the hall of fame, in a
        way that inserting a new individual in a full hall of fame will not
        remove the worst individual to maintain a constant size.

        :param item: The individual with a fitness attribute to insert in the
                     hall of fame.
        """
        i = bisect_right(self.keys, item.fitness)
        self.items.insert(len(self) - i, item)
        self.keys.insert(i, item.fitness)

    def remove(self, index: int):
        """
        Remove the specified *index* from the hall of fame.

        :param index: An integer giving which item to remove.
        """
        del self.keys[len(self) - (index % len(self) + 1)]
        del self.items[index]

    def clear(self):
        """Clear the hall of fame."""
        del self.items[:]
        del self.keys[:]

    def __len__(self):
        return len(self.items)

    def __getitem__(self, i):
        return self.items[i]

    def __iter__(self):
        return iter(self.items)

    def __reversed__(self):
        return reversed(self.items)

    def __str__(self):
        return str(self.items)


class ParetoFront(HallOfFame):
    """
    The Pareto front hall of fame contains all the non-dominated individuals
    that ever lived in the population. That means that the Pareto front hall of
    fame can contain an infinity of different individuals.

    :param similar: A function that tells the Pareto front whether or not two
                    individuals are similar, optional.

    The size of the front may become very large if it is used for example on
    a continuous function with a continuous domain. In order to limit the number
    of individuals, it is possible to specify a similarity function that will
    return :data:`True` if the genotype of two individuals are similar. In that
    case only one of the two individuals will be added to the hall of fame. By
    default the similarity function is :func:`operator.eq`.

    ParetoFront also supports hard-limiting maxsize, in which cases the element
    with worst primary metric is removed. By default ParetoFront is unbounded.

    Since, the Pareto front hall of fame inherits from the :class:`HallOfFame`,
    it is sorted lexicographically at every moment.
    """

    def __init__(self, maxsize: Optional[int] = None, similar: Callable = eq):
        HallOfFame.__init__(self, maxsize, similar)

    def update(self, population: PopulationT):
        """
        Update the Pareto front hall of fame with the *population* by adding
        the individuals from the population that are not dominated by the hall
        of fame. If any individual in the hall of fame is dominated it is
        removed.

        :param population: A list of individual with a fitness attribute to
                           update the hall of fame with.
        """
        for ind in population:
            is_dominated = False
            dominates_one = False
            has_twin = False
            to_remove = []
            for i, hof_member in enumerate(self):  # hall of fame member
                if not dominates_one and hof_member.fitness.dominates(ind.fitness):
                    is_dominated = True
                    break
                elif ind.fitness.dominates(hof_member.fitness):
                    dominates_one = True
                    to_remove.append(i)
                elif ind.fitness == hof_member.fitness and self.similar(ind, hof_member):
                    has_twin = True
                    break

            for i in reversed(to_remove):  # Remove the dominated hofer
                self.remove(i)
            if not is_dominated and not has_twin:
                if len(self) >= self.maxsize > 0:
                    self.remove(-1)
                self.insert(ind)



from typing import Any, Dict

from golem.core.dag.linked_graph_node import LinkedGraphNode
from .. import any_to_json


def graph_node_to_json(obj: LinkedGraphNode) -> Dict[str, Any]:
    """
    Uses regular serialization but excludes "_operator" field to rid of circular references
    """
    encoded = {
        k: v
        for k, v in any_to_json(obj).items()
        if k not in ['_operator', '_fitted_operation', '_node_data', '_parameters']
    }
    if 'name' in encoded['content']:
        encoded['content']['name'] = str(encoded['content']['name'])
    if encoded['_nodes_from']:
        encoded['_nodes_from'] = [
            node.uid
            for node in encoded['_nodes_from']
        ]
    return encoded



from datetime import timedelta
from functools import partial
from typing import Callable, Optional

from hyperopt import tpe, fmin, space_eval

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.objective import ObjectiveFunction
from golem.core.tuning.hyperopt_tuner import HyperoptTuner, get_node_parameters_for_hyperopt
from golem.core.tuning.search_space import SearchSpace
from golem.core.tuning.tuner_interface import DomainGraphForTune


class SequentialTuner(HyperoptTuner):
    """
    Class for hyperparameters optimization for all nodes sequentially
    """

    def __init__(self, objective_evaluate: ObjectiveFunction,
                 search_space: SearchSpace,
                 adapter: Optional[BaseOptimizationAdapter] = None,
                 iterations: int = 100,
                 early_stopping_rounds: Optional[int] = None,
                 timeout: timedelta = timedelta(minutes=5),
                 n_jobs: int = -1,
                 deviation: float = 0.05,
                 algo: Callable = tpe.suggest,
                 inverse_node_order: bool = False):
        super().__init__(objective_evaluate,
                         search_space,
                         adapter,
                         iterations,
                         early_stopping_rounds, timeout,
                         n_jobs,
                         deviation,
                         algo)

        self.inverse_node_order = inverse_node_order

    def tune(self, graph: DomainGraphForTune) -> DomainGraphForTune:
        """ Method for hyperparameters tuning on the entire graph

        Args:
            graph: graph which hyperparameters will be tuned
        """
        graph = self.adapter.adapt(graph)

        # Check source metrics for data
        self.init_check(graph)

        # Calculate amount of iterations we can apply per node
        nodes_amount = graph.length
        iterations_per_node = round(self.iterations / nodes_amount)
        iterations_per_node = int(iterations_per_node)
        if iterations_per_node == 0:
            iterations_per_node = 1

        # Calculate amount of seconds we can apply per node
        if self.max_seconds is not None:
            seconds_per_node = round(self.max_seconds / nodes_amount)
            seconds_per_node = int(seconds_per_node)
        else:
            seconds_per_node = None

        # Tuning performed sequentially for every node - so get ids of nodes
        nodes_ids = self.get_nodes_order(nodes_number=nodes_amount)
        for node_id in nodes_ids:
            node = graph.nodes[node_id]
            operation_name = node.name

            # Get node's parameters to optimize
            node_params = get_node_parameters_for_hyperopt(self.search_space, node_id, operation_name)

            if not node_params:
                self.log.info(f'"{operation_name}" operation has no parameters to optimize')
            else:
                # Apply tuning for current node
                self._optimize_node(node_id=node_id,
                                    graph=graph,
                                    node_params=node_params,
                                    iterations_per_node=iterations_per_node,
                                    seconds_per_node=seconds_per_node)

        # Validate if optimisation did well
        final_graph = self.final_check(graph)

        self.was_tuned = True
        final_graph = self.adapter.restore(final_graph)

        return final_graph

    def get_nodes_order(self, nodes_number: int) -> range:
        """ Method returns list with indices of nodes in the graph

        Args:
            nodes_number: number of nodes to get
        """

        if self.inverse_node_order is True:
            # From source data to output
            nodes_ids = range(nodes_number - 1, -1, -1)
        else:
            # From output to source data
            nodes_ids = range(0, nodes_number)

        return nodes_ids

    def tune_node(self, graph: DomainGraphForTune, node_index: int) -> DomainGraphForTune:
        """ Method for hyperparameters tuning for particular node

        Args:
            graph: graph which contains a node to be tuned
            node_index: Index of the node to tune

        Returns:
            Graph with tuned parameters in node with specified index
        """
        graph = self.adapter.adapt(graph)

        self.init_check(graph)

        node = graph.nodes[node_index]
        operation_name = node.name

        # Get node's parameters to optimize
        node_params = get_node_parameters_for_hyperopt(self.search_space,
                                                       node_id=node_index,
                                                       operation_name=operation_name)

        if not node_params:
            self._stop_tuning_with_message(f'"{operation_name}" operation has no parameters to optimize')
        else:
            # Apply tuning for current node
            self._optimize_node(graph=graph,
                                node_id=node_index,
                                node_params=node_params,
                                iterations_per_node=self.iterations,
                                seconds_per_node=self.max_seconds,
                                )
            self.was_tuned = True

        # Validation is the optimization do well
        final_graph = self.final_check(graph)
        final_graph = self.adapter.restore(final_graph)
        return final_graph

    def _optimize_node(self, graph: OptGraph,
                       node_id: int,
                       node_params: dict,
                       iterations_per_node: int,
                       seconds_per_node: int) -> OptGraph:
        """
        Method for node optimization

        Args:
            graph: Graph which node is optimized
            node_id: id of the current node in the graph
            node_params: dictionary with parameters for node
            iterations_per_node: amount of iterations to produce
            seconds_per_node: amount of seconds to produce

        Returns:
            updated graph with tuned parameters in particular node
        """
        best_parameters = fmin(partial(self._objective, graph=graph, node_id=node_id),
                               node_params,
                               algo=self.algo,
                               max_evals=iterations_per_node,
                               early_stop_fn=self.early_stop_fn,
                               timeout=seconds_per_node)

        best_parameters = space_eval(space=node_params, hp_assignment=best_parameters)

        # Set best params for this node in the graph
        graph = self.set_arg_node(graph=graph, node_id=node_id, node_params=best_parameters)
        return graph

    def _objective(self, node_params: dict, graph: OptGraph, node_id: int) -> float:
        """ Objective function for minimization problem

        Args:
            node_params: dictionary with parameters for node
            graph: graph to evaluate
            node_id: id of the node to which parameters should be assigned

        Returns:
            value of objective function
        """

        # Set hyperparameters for node
        graph = self.set_arg_node(graph=graph, node_id=node_id, node_params=node_params)

        metric_value = self.get_metric_value(graph=graph)
        return metric_value



from datetime import timedelta, datetime
from itertools import zip_longest
from typing import Optional, Callable, Dict, Sequence

import networkx as nx
import numpy as np
import zss
from networkx import graph_edit_distance, is_tree

from examples.synthetic_graph_evolution.generators import generate_labeled_graph
from golem.core.optimisers.optimization_parameters import GraphRequirements
from golem.metrics.graph_metrics import min_max
from libs.netcomp import edit_distance


def _label_dist(label1: str, label2: str) -> int:
    return int(label1 != label2)


def tree_edit_dist(target_graph: nx.DiGraph, graph: nx.DiGraph) -> float:
    """Compares nodes by their `name` (if present) or `uid` attribute.
    Nodes with the same name/id are considered the same."""
    if not (is_tree(target_graph) and is_tree(graph)):
        raise ValueError('Both target graphs must be trees')
    target_tree_root = _nx_to_zss_tree(target_graph)
    cmp_tree_root = _nx_to_zss_tree(graph)
    dist = zss.simple_distance(target_tree_root, cmp_tree_root, label_dist=_label_dist)
    return dist


def graph_size(target_graph: nx.DiGraph, graph: nx.DiGraph) -> int:
    return abs(target_graph.number_of_nodes() - graph.number_of_nodes())


def _nx_to_zss_tree(graph: nx.DiGraph) -> zss.Node:
    # Root is the node without successors
    root = _get_root_node(graph)
    # that's why we first reverse the tree to get proper DFS traverse
    tree = graph.reverse()
    # Add nodes with appropriate labels for comparison
    nodes_dict = {}
    for node_id, node_data in tree.nodes(data=True):
        label = node_data.get('name', node_id)
        nodes_dict[node_id] = zss.Node(label)
    # Add edges
    for edge in tree.edges():
        nodes_dict[edge[0]].addkid(nodes_dict[edge[1]])
    return nodes_dict[root]


def _get_root_node(nxgraph: nx.DiGraph) -> Sequence:
    source = [n for (n, d) in nxgraph.out_degree() if d == 0][0]
    return source


def get_edit_dist_metric(target_graph: nx.DiGraph,
                         timeout=timedelta(seconds=60),
                         upper_bound: Optional[int] = None,
                         requirements: Optional[GraphRequirements] = None,
                         ) -> Callable[[nx.DiGraph], float]:
    def node_match(node_content_1: Dict, node_content_2: Dict) -> bool:
        operations_do_match = node_content_1.get('name') == node_content_2.get('name')
        return True or operations_do_match

    if requirements:
        upper_bound = upper_bound or int(np.sqrt(requirements.max_depth * requirements.max_arity)),
        timeout = timeout or requirements.max_graph_fit_time

    def metric(graph: nx.DiGraph) -> float:
        ged = graph_edit_distance(target_graph, graph,
                                  node_match=node_match,
                                  upper_bound=upper_bound,
                                  timeout=timeout.total_seconds() if timeout else None,
                                  )
        return float(ged) or upper_bound

    return metric


def matrix_edit_dist(target_graph: nx.DiGraph, graph: nx.DiGraph) -> float:
    target_adj = nx.adjacency_matrix(target_graph)
    adj = nx.adjacency_matrix(graph)
    nmin, nmax = min_max(target_adj.shape[0], adj.shape[0])
    if nmin != nmax:
        shape = (nmax, nmax)
        target_adj.resize(shape)
        adj.resize(shape)
    value = edit_distance(target_adj, adj)
    return value


def try_tree_edit_distance(sizes1=None, sizes2=None, node_types=None,
                           print_trees=False):
    if not sizes1:
        sizes1 = list(range(5, 100, 5))
    if not sizes2:
        sizes2 = sizes1
    if not node_types:
        node_types = ['X']

    for i, (n1, n2) in enumerate(zip_longest(sizes1, sizes2)):
        g1 = generate_labeled_graph('tree', n1, node_types)
        g2 = generate_labeled_graph('tree', n2, node_types)

        start_time = datetime.now()
        dist = tree_edit_dist(g1, g2)
        duration = datetime.now() - start_time

        print(f'iter {i} with sizes={(n1, n2)} dist={dist}, '
              f't={duration.total_seconds():.3f}s')
        if print_trees:
            print(nx.forest_str(g1))
            print(nx.forest_str(g2))


if __name__ == "__main__":
    try_tree_edit_distance(print_trees=False, node_types=list('XYZWQ'))



import pathlib
import timeit
from datetime import datetime
from typing import Optional, Tuple

from golem.core.adapter import BaseOptimizationAdapter
from golem.core.log import Log
from golem.core.optimisers.genetic.evaluation import DelegateEvaluator, SequentialDispatcher
from golem.core.optimisers.graph import OptGraph
from golem.core.optimisers.meta.surrogate_model import SurrogateModel, RandomValuesSurrogateModel
from golem.core.optimisers.objective.objective import to_fitness, GraphFunction
from golem.core.optimisers.opt_history_objects.individual import GraphEvalResult


class SurrogateDispatcher(SequentialDispatcher):
    """Evaluates objective function with surrogate model.
        Usage: call `dispatch(objective_function)` to get evaluation function.
        Additionally, we need to pass surrogate_model object
    """

    def __init__(self,
                 adapter: BaseOptimizationAdapter,
                 n_jobs: int = 1,
                 graph_cleanup_fn: Optional[GraphFunction] = None,
                 delegate_evaluator: Optional[DelegateEvaluator] = None,
                 surrogate_model: SurrogateModel = RandomValuesSurrogateModel()):
        super().__init__(adapter, n_jobs, graph_cleanup_fn, delegate_evaluator)
        self._n_jobs = 1
        self.surrogate_model = surrogate_model

    def evaluate_single(self, graph: OptGraph, uid_of_individual: str, with_time_limit: bool = True,
                        cache_key: Optional[str] = None,
                        logs_initializer: Optional[Tuple[int, pathlib.Path]] = None) -> GraphEvalResult:
        graph = self.evaluation_cache.get(cache_key, graph)
        if logs_initializer is not None:
            # in case of multiprocessing run
            Log.setup_in_mp(*logs_initializer)

        start_time = timeit.default_timer()
        fitness = to_fitness(self.surrogate_model(graph, objective=self._objective_eval))
        end_time = timeit.default_timer()

        eval_res = GraphEvalResult(
            uid_of_individual=uid_of_individual, fitness=fitness, graph=graph, metadata={
                'computation_time_in_seconds': end_time - start_time,
                'evaluation_time_iso': datetime.now().isoformat(),
                'surrogate_evaluation': True
            }
        )
        return eval_res



import itertools
import numpy as np
from scipy.linalg import toeplitz
from scipy.stats import wasserstein_distance


def emd(x, y, distance_scaling=1.0):
    """Earth's mover distance (or Wasserstein metric)
     between 2 probability distributions."""
    support_size = max(len(x), len(y))
    d_mat = toeplitz(range(support_size)).astype(np.float64)
    distance_mat = d_mat / distance_scaling

    # convert histogram values x and y to float, and make them equal len
    x = x.astype(np.float64)
    y = y.astype(np.float64)
    if len(x) < len(y):
        x = np.hstack((x, [0.0] * (support_size - len(x))))
    elif len(y) < len(x):
        y = np.hstack((y, [0.0] * (support_size - len(y))))

    emd_value = wasserstein_distance(x, y, distance_mat)
    return emd_value


def l2(x, y):
    dist = np.linalg.norm(x - y, 2)
    return dist


def gaussian_emd(x, y, sigma=1.0, distance_scaling=1.0):
    ''' Gaussian kernel with squared distance in exponential term replaced by EMD
    Args:
      x, y: 1D pmf of two distributions with the same support
      sigma: standard deviation
    '''
    emd_val = emd(x, y, distance_scaling=distance_scaling)
    return np.exp(-emd_val ** 2 / (2 * sigma ** 2))


def gaussian(x, y, sigma=1.0):
    dist = np.linalg.norm(x - y, 2)
    return np.exp(-dist ** 2 / (2 * sigma ** 2))


def discrepancy(samples1, samples2, kernel, *args, **kwargs) -> float:
    d = sum(kernel(s1, s2, *args, **kwargs) for s1, s2 in itertools.product(samples1, samples2))
    d /= len(samples1) * len(samples2)
    return d


def compute_mmd(samples1, samples2, kernel=gaussian_emd, normalize=True, *args, **kwargs) -> float:
    ''' MMD between two samples
    '''
    # normalize histograms into pmf
    if normalize:
        samples1 = [s1 / np.sum(s1) for s1 in samples1]
        samples2 = [s2 / np.sum(s2) for s2 in samples2]
    return discrepancy(samples1, samples1, kernel, *args, **kwargs) + \
           discrepancy(samples2, samples2, kernel, *args, **kwargs) - \
           2 * discrepancy(samples1, samples2, kernel, *args, **kwargs)


def compute_emd(samples1, samples2, kernel, normalize=True, *args, **kwargs) -> float:
    ''' EMD between average of two samples
    '''
    if normalize:
        samples1 = [np.mean(samples1)]
        samples2 = [np.mean(samples2)]
    return discrepancy(samples1, samples2, kernel, *args, **kwargs)


def test_mmd_compute():
    s1 = np.array([0.2, 0.8])
    s2 = np.array([0.3, 0.7])
    samples1 = [s1, s2]

    s3 = np.array([0.25, 0.75])
    s4 = np.array([0.35, 0.65])
    samples2 = [s3, s4]

    s5 = np.array([0.8, 0.2])
    s6 = np.array([0.7, 0.3])
    samples3 = [s5, s6]

    print('between samples1 and samples2: ', compute_mmd(samples1, samples2, kernel=gaussian_emd, sigma=1.0))
    print('between samples1 and samples3: ', compute_mmd(samples1, samples3, kernel=gaussian_emd, sigma=1.0))


if __name__ == '__main__':
    test_mmd_compute()



from typing import Tuple

from golem.core.optimisers.genetic.gp_params import GPAlgorithmParameters
from golem.core.optimisers.genetic.operators.inheritance import GeneticSchemeTypesEnum
from golem.core.optimisers.genetic.operators.operator import PopulationT
from golem.core.optimisers.genetic.parameters.mutation_prob import AdaptiveMutationProb
from golem.core.optimisers.genetic.parameters.parameter import AdaptiveParameter, ConstParameter

VariationOperatorProb = AdaptiveParameter[Tuple[float, float]]


class AdaptiveVariationProb(VariationOperatorProb):
    """Adaptive parameter for variation operators.
    Specifies mutation and crossover probabilities."""

    def __init__(self, mutation_prob: float = 0.5, crossover_prob: float = 0.5):
        self._mutation_prob_param = AdaptiveMutationProb(mutation_prob)
        self._mutation_prob = self._mutation_prob_param.initial
        self._crossover_prob_init = crossover_prob
        self._crossover_prob = crossover_prob

    @property
    def mutation_prob(self):
        return self._mutation_prob

    @property
    def crossover_prob(self):
        return self._crossover_prob

    @property
    def initial(self) -> Tuple[float, float]:
        return self._mutation_prob_param.initial, self._crossover_prob_init

    def next(self, population: PopulationT) -> Tuple[float, float]:
        self._mutation_prob = self._mutation_prob_param.next(population)
        self._crossover_prob = 1. - self._mutation_prob
        return self._mutation_prob, self._crossover_prob


def init_adaptive_operators_prob(parameters: GPAlgorithmParameters) -> VariationOperatorProb:
    """Returns static or adaptive parameter for mutation & crossover probabilities depending on genetic type scheme."""
    if parameters.genetic_scheme_type == GeneticSchemeTypesEnum.parameter_free:
        operators_prob = AdaptiveVariationProb()
    else:
        operators_prob = ConstParameter((parameters.mutation_prob, parameters.crossover_prob))
    return operators_prob
